# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii                      import *
from torii.hdl.rec              import *
from torii.sim                  import *
from torii.lib.soc.wishbone.bus import *
from torii.lib.soc.memory       import MemoryMap

from ....utils                  import ToriiTestSuiteCase

class InterfaceTestCase(ToriiTestSuiteCase):
	def test_simple(self):
		iface = Interface(addr_width = 32, data_width = 8)
		self.assertEqual(iface.addr_width, 32)
		self.assertEqual(iface.data_width, 8)
		self.assertEqual(iface.granularity, 8)
		self.assertEqual(iface.layout, Layout.cast([
			('adr',   32, DIR_FANOUT),
			('dat_w', 8,  DIR_FANOUT),
			('dat_r', 8,  DIR_FANIN),
			('sel',   1,  DIR_FANOUT),
			('cyc',   1,  DIR_FANOUT),
			('stb',   1,  DIR_FANOUT),
			('we',    1,  DIR_FANOUT),
			('ack',   1,  DIR_FANIN),
		]))

	def test_granularity(self):
		iface = Interface(addr_width = 30, data_width = 32, granularity = 8)
		self.assertEqual(iface.addr_width, 30)
		self.assertEqual(iface.data_width, 32)
		self.assertEqual(iface.granularity, 8)
		self.assertEqual(iface.layout, Layout.cast([
			('adr',   30, DIR_FANOUT),
			('dat_w', 32, DIR_FANOUT),
			('dat_r', 32, DIR_FANIN),
			('sel',   4,  DIR_FANOUT),
			('cyc',   1,  DIR_FANOUT),
			('stb',   1,  DIR_FANOUT),
			('we',    1,  DIR_FANOUT),
			('ack',   1,  DIR_FANIN),
		]))

	def test_features(self):
		iface = Interface(
			addr_width = 32, data_width = 32,
			features = {'rty', 'err', 'stall', 'lock', 'cti', 'bte'}
		)
		self.assertEqual(iface.layout, Layout.cast([
			('adr',   32, DIR_FANOUT),
			('dat_w', 32, DIR_FANOUT),
			('dat_r', 32, DIR_FANIN),
			('sel',   1,  DIR_FANOUT),
			('cyc',   1,  DIR_FANOUT),
			('stb',   1,  DIR_FANOUT),
			('we',    1,  DIR_FANOUT),
			('ack',   1,  DIR_FANIN),
			('err',   1,  DIR_FANIN),
			('rty',   1,  DIR_FANIN),
			('stall', 1,  DIR_FANIN),
			('lock',  1,  DIR_FANOUT),
			('cti',   CycleType,    DIR_FANOUT),
			('bte',   BurstTypeExt, DIR_FANOUT),
		]))

	def test_wrong_addr_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Address width must be a non-negative integer, not -1'
		):
			Interface(addr_width = -1, data_width = 8)

	def test_wrong_data_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Data width must be one of 8, 16, 32, 64, not 7'
		):
			Interface(addr_width = 0, data_width = 7)

	def test_wrong_granularity(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Granularity must be one of 8, 16, 32, 64, not 7'
		):
			Interface(addr_width = 0, data_width = 32, granularity = 7)

	def test_wrong_granularity_wide(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Granularity 32 may not be greater than data width 8'
		):
			Interface(addr_width = 0, data_width = 8, granularity = 32)

	def test_wrong_features(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Optional signal\(s\) \'foo\' are not supported'
		):
			Interface(addr_width = 0, data_width = 8, features = {'foo'})

	def test_get_map_wrong(self):
		iface = Interface(addr_width = 0, data_width = 8)
		with self.assertRaisesRegex(
			NotImplementedError,
			r'Bus interface \(rec iface adr dat_w dat_r sel cyc stb we ack\) does '
			r'not have a memory map'
		):
			iface.memory_map

	def test_get_map_frozen(self):
		iface = Interface(addr_width = 1, data_width = 8)
		iface.memory_map = MemoryMap(addr_width = 1, data_width = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has been frozen\. Address width cannot be extended '
			r'further'
		):
			iface.memory_map.addr_width = 2

	def test_set_map_wrong(self):
		iface = Interface(addr_width = 0, data_width = 8)
		with self.assertRaisesRegex(
			TypeError,
			r'Memory map must be an instance of MemoryMap, not \'foo\''
		):
			iface.memory_map = 'foo'

	def test_set_map_wrong_data_width(self):
		iface = Interface(addr_width = 30, data_width = 32, granularity = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has data width 32, which is not the same as bus '
			r'interface granularity 8'
		):
			iface.memory_map = MemoryMap(addr_width = 32, data_width = 32)

	def test_set_map_wrong_addr_width(self):
		iface = Interface(addr_width = 30, data_width = 32, granularity = 8)
		with self.assertRaisesRegex(
			ValueError,
			r'Memory map has address width 30, which is not the same as bus '
			r'interface address width 32 \(30 address bits \+ 2 granularity bits\)'
		):
			iface.memory_map = MemoryMap(addr_width = 30, data_width = 8)


class DecoderTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.dut = Decoder(addr_width = 31, data_width = 32, granularity = 16)

	def test_add_align_to(self):
		sub_1 = Interface(addr_width = 15, data_width = 32, granularity = 16)
		sub_1.memory_map = MemoryMap(addr_width = 16, data_width = 16)
		sub_2 = Interface(addr_width = 15, data_width = 32, granularity = 16)
		sub_2.memory_map = MemoryMap(addr_width = 16, data_width = 16)
		self.assertEqual(self.dut.add(sub_1), (0x00000000, 0x00010000, 1))
		self.assertEqual(self.dut.align_to(18), 0x000040000)
		self.assertEqual(self.dut.align_to(alignment = 18), 0x000040000)
		self.assertEqual(self.dut.add(sub_2), (0x00040000, 0x00050000, 1))

	def test_add_extend(self):
		sub = Interface(addr_width = 31, data_width = 32, granularity = 16)
		sub.memory_map = MemoryMap(addr_width = 32, data_width = 16)
		self.assertEqual(self.dut.add(sub, addr = 1, extend = True), (1, 0x100000001, 1))
		self.assertEqual(self.dut.bus.addr_width, 32)

	def test_add_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Subordinate bus must be an instance of wishbone\.Interface, not \'foo\''
		):
			self.dut.add(sub_bus = 'foo')

	def test_add_wrong_granularity(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Subordinate bus has granularity 32, which is greater than '
			r'the decoder granularity 16'
		):
			self.dut.add(Interface(addr_width = 15, data_width = 32, granularity = 32))

	def test_add_wrong_width_dense(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Subordinate bus has data width 16, which is not the same as decoder '
			r'data width 32 \(required for dense address translation\)'
		):
			self.dut.add(Interface(addr_width = 15, data_width = 16, granularity = 16))

	def test_add_wrong_granularity_sparse(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Subordinate bus has data width 64, which is not the same as subordinate '
			r'bus granularity 16 \(required for sparse address translation\)'
		):
			self.dut.add(
				Interface(
					addr_width = 15, data_width = 64, granularity = 16
				),
				sparse = True
			)

	def test_add_wrong_optional_output(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Subordinate bus has optional output \'err\', but the decoder does '
			r'not have a corresponding input'
		):
			self.dut.add(
				Interface(
					addr_width = 15, data_width = 32, granularity = 16, features = {'err'}
				)
			)

	def test_add_wrong_out_of_bounds(self):
		sub = Interface(addr_width = 31, data_width = 32, granularity = 16)
		sub.memory_map = MemoryMap(addr_width = 32, data_width = 16)
		with self.assertRaisesRegex(
			ValueError,
			r'Address range 0x1\.\.0x100000001 out of bounds for memory map spanning '
			r'range 0x0\.\.0x100000000 \(32 address bits\)'
		):
			self.dut.add(sub, addr = 1)


class DecoderSimulationTestCase(ToriiTestSuiteCase):
	def test_simple(self):
		dut = Decoder(
			addr_width = 30, data_width = 32, granularity = 8,
			features = {'err', 'rty', 'stall', 'lock', 'cti', 'bte'}
		)
		sub_1 = Interface(addr_width = 14, data_width = 32, granularity = 8)
		sub_1.memory_map = MemoryMap(addr_width = 16, data_width = 8)
		dut.add(sub_1, addr = 0x10000)
		sub_2 = Interface(
			addr_width = 14, data_width = 32, granularity = 8,
			features = {'err', 'rty', 'stall', 'lock', 'cti', 'bte'}
		)
		sub_2.memory_map = MemoryMap(addr_width = 16, data_width = 8)
		dut.add(sub_2)

		def sim_test():
			yield dut.bus.adr.eq(0x10400 >> 2)
			yield dut.bus.cyc.eq(1)
			yield dut.bus.stb.eq(1)
			yield dut.bus.sel.eq(0b11)
			yield dut.bus.dat_w.eq(0x12345678)
			yield dut.bus.lock.eq(1)
			yield dut.bus.cti.eq(CycleType.INCR_BURST)
			yield dut.bus.bte.eq(BurstTypeExt.WRAP_4)
			yield sub_1.ack.eq(1)
			yield sub_1.dat_r.eq(0xabcdef01)
			yield sub_2.dat_r.eq(0x5678abcd)
			yield Delay(1e-6)
			self.assertEqual((yield sub_1.adr), 0x400 >> 2)
			self.assertEqual((yield sub_1.cyc), 1)
			self.assertEqual((yield sub_2.cyc), 0)
			self.assertEqual((yield sub_1.stb), 1)
			self.assertEqual((yield sub_1.sel), 0b11)
			self.assertEqual((yield sub_1.dat_w), 0x12345678)
			self.assertEqual((yield dut.bus.ack), 1)
			self.assertEqual((yield dut.bus.err), 0)
			self.assertEqual((yield dut.bus.rty), 0)
			self.assertEqual((yield dut.bus.dat_r), 0xabcdef01)

			yield dut.bus.adr.eq(0x20400 >> 2)
			yield sub_1.ack.eq(0)
			yield sub_2.err.eq(1)
			yield sub_2.rty.eq(1)
			yield sub_2.stall.eq(1)
			yield Delay(1e-6)
			self.assertEqual((yield sub_2.adr), 0x400 >> 2)
			self.assertEqual((yield sub_1.cyc), 0)
			self.assertEqual((yield sub_2.cyc), 1)
			self.assertEqual((yield sub_1.stb), 1)
			self.assertEqual((yield sub_1.sel), 0b11)
			self.assertEqual((yield sub_1.dat_w), 0x12345678)
			self.assertEqual((yield sub_2.lock), 1)
			self.assertEqual((yield sub_2.cti), CycleType.INCR_BURST.value)
			self.assertEqual((yield sub_2.bte), BurstTypeExt.WRAP_4.value)
			self.assertEqual((yield dut.bus.ack), 0)
			self.assertEqual((yield dut.bus.err), 1)
			self.assertEqual((yield dut.bus.rty), 1)
			self.assertEqual((yield dut.bus.stall), 1)
			self.assertEqual((yield dut.bus.dat_r), 0x5678abcd)

		sim = Simulator(dut)
		sim.add_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()

	def test_addr_translate(self):
		class AddressLoopback(Elaboratable):
			def __init__(self, **kwargs):
				self.bus = Interface(**kwargs)

			def elaborate(self, platform):
				m = Module()

				for index, sel_bit in enumerate(self.bus.sel):
					with m.If(sel_bit):
						segment = self.bus.dat_r.word_select(index, self.bus.granularity)
						m.d.comb += segment.eq(self.bus.adr + index)

				return m

		dut = Decoder(addr_width = 20, data_width = 32, granularity = 16)
		loop_1 = AddressLoopback(addr_width = 7, data_width = 32, granularity = 16)
		loop_1.bus.memory_map = MemoryMap(addr_width = 8, data_width = 16)
		self.assertEqual(
			dut.add(loop_1.bus, addr = 0x10000),
			(0x10000, 0x10100, 1)
		)
		loop_2 = AddressLoopback(addr_width = 6, data_width = 32, granularity = 8)
		loop_2.bus.memory_map = MemoryMap(addr_width = 8, data_width = 8)
		self.assertEqual(
			dut.add(loop_2.bus, addr = 0x20000),
			(0x20000, 0x20080, 2)
		)
		loop_3 = AddressLoopback(addr_width = 8, data_width = 16, granularity = 16)
		loop_3.bus.memory_map = MemoryMap(addr_width = 8, data_width = 16)
		self.assertEqual(
			dut.add(loop_3.bus, addr = 0x30000, sparse = True),
			(0x30000, 0x30100, 1)
		)
		loop_4 = AddressLoopback(addr_width = 8, data_width = 8,  granularity = 8)
		loop_4.bus.memory_map = MemoryMap(addr_width = 8, data_width = 8)
		self.assertEqual(
			dut.add(loop_4.bus, addr = 0x40000, sparse = True),
			(0x40000, 0x40100, 1)
		)

		def sim_test():
			yield dut.bus.cyc.eq(1)

			yield dut.bus.adr.eq(0x10010 >> 1)

			yield dut.bus.sel.eq(0b11)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x00090008)

			yield dut.bus.sel.eq(0b01)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x00000008)

			yield dut.bus.sel.eq(0b10)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x00090000)

			yield dut.bus.adr.eq(0x20010 >> 1)

			yield dut.bus.sel.eq(0b11)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x13121110)

			yield dut.bus.sel.eq(0b01)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x00001110)

			yield dut.bus.sel.eq(0b10)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x13120000)

			yield dut.bus.adr.eq(0x30010 >> 1)

			yield dut.bus.sel.eq(0b11)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x0008)

			yield dut.bus.sel.eq(0b01)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x0008)

			yield dut.bus.sel.eq(0b10)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x0000)

			yield dut.bus.adr.eq(0x30012 >> 1)

			yield dut.bus.sel.eq(0b11)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x0009)

			yield dut.bus.adr.eq(0x40010 >> 1)

			yield dut.bus.sel.eq(0b11)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x08)

			yield dut.bus.sel.eq(0b01)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x08)

			yield dut.bus.sel.eq(0b10)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x00)

			yield dut.bus.adr.eq(0x40012 >> 1)

			yield dut.bus.sel.eq(0b11)
			yield Delay(1e-6)
			self.assertEqual((yield dut.bus.dat_r), 0x09)

		m = Module()
		m.submodules += dut, loop_1, loop_2, loop_3, loop_4
		sim = Simulator(m)
		sim.add_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()

	def test_coarse_granularity(self):
		dut = Decoder(addr_width = 3, data_width = 32)
		sub = Interface(addr_width = 2, data_width = 32)
		sub.memory_map = MemoryMap(addr_width = 2, data_width = 32)
		dut.add(sub)

		def sim_test():
			yield dut.bus.cyc.eq(1)

			yield dut.bus.adr.eq(0x0)
			yield Delay(1e-6)
			self.assertEqual((yield sub.cyc), 1)

			yield dut.bus.adr.eq(0x4)
			yield Delay(1e-6)
			self.assertEqual((yield sub.cyc), 0)

		sim = Simulator(dut)
		sim.add_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()


class ArbiterTestCase(ToriiTestSuiteCase):
	def setUp(self):
		self.dut = Arbiter(
			addr_width = 31, data_width = 32, granularity = 16,
			features = {'err'}
		)

	def test_add_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'Initiator bus must be an instance of wishbone\.Interface, not \'foo\''
		):
			self.dut.add(intr_bus = 'foo')

	def test_add_wrong_addr_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Initiator bus has address width 15, which is not the same as arbiter '
			r'address width 31'
		):
			self.dut.add(Interface(addr_width = 15, data_width = 32, granularity = 16))

	def test_add_wrong_granularity(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Initiator bus has granularity 8, which is lesser than '
			r'the arbiter granularity 16'
		):
			self.dut.add(Interface(addr_width = 31, data_width = 32, granularity = 8))

	def test_add_wrong_data_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Initiator bus has data width 16, which is not the same as arbiter '
			r'data width 32'
		):
			self.dut.add(Interface(addr_width = 31, data_width = 16, granularity = 16))

	def test_add_wrong_optional_output(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Arbiter has optional output \'err\', but the initiator bus does '
			r'not have a corresponding input'
		):
			self.dut.add(Interface(addr_width = 31, data_width = 32, granularity = 16))


class ArbiterSimulationTestCase(ToriiTestSuiteCase):
	def test_simple(self):
		dut = Arbiter(
			addr_width = 30, data_width = 32, granularity = 8,
			features = {'err', 'rty', 'stall', 'lock', 'cti', 'bte'}
		)
		intr_1 = Interface(
			addr_width = 30, data_width = 32, granularity = 8,
			features = {'err', 'rty'}
		)
		dut.add(intr_1)
		intr_2 = Interface(
			addr_width = 30, data_width = 32, granularity = 16,
			features = {'err', 'rty', 'stall', 'lock', 'cti', 'bte'}
		)
		dut.add(intr_2)

		def sim_test():
			yield intr_1.adr.eq(0x7ffffffc >> 2)
			yield intr_1.cyc.eq(1)
			yield intr_1.stb.eq(1)
			yield intr_1.sel.eq(0b1111)
			yield intr_1.we.eq(1)
			yield intr_1.dat_w.eq(0x12345678)
			yield dut.bus.dat_r.eq(0xabcdef01)
			yield dut.bus.ack.eq(1)
			yield dut.bus.err.eq(1)
			yield dut.bus.rty.eq(1)
			yield Delay(1e-7)
			self.assertEqual((yield dut.bus.adr), 0x7ffffffc >> 2)
			self.assertEqual((yield dut.bus.cyc), 1)
			self.assertEqual((yield dut.bus.stb), 1)
			self.assertEqual((yield dut.bus.sel), 0b1111)
			self.assertEqual((yield dut.bus.we), 1)
			self.assertEqual((yield dut.bus.dat_w), 0x12345678)
			self.assertEqual((yield dut.bus.lock), 0)
			self.assertEqual((yield dut.bus.cti), CycleType.CLASSIC.value)
			self.assertEqual((yield dut.bus.bte), BurstTypeExt.LINEAR.value)
			self.assertEqual((yield intr_1.dat_r), 0xabcdef01)
			self.assertEqual((yield intr_1.ack), 1)
			self.assertEqual((yield intr_1.err), 1)
			self.assertEqual((yield intr_1.rty), 1)

			yield intr_1.cyc.eq(0)
			yield intr_2.adr.eq(0xe0000000 >> 2)
			yield intr_2.cyc.eq(1)
			yield intr_2.stb.eq(1)
			yield intr_2.sel.eq(0b10)
			yield intr_2.we.eq(1)
			yield intr_2.dat_w.eq(0x43218765)
			yield intr_2.lock.eq(0)
			yield intr_2.cti.eq(CycleType.INCR_BURST)
			yield intr_2.bte.eq(BurstTypeExt.WRAP_4)
			yield Tick()

			yield dut.bus.stall.eq(0)
			yield Delay(1e-7)
			self.assertEqual((yield dut.bus.adr), 0xe0000000 >> 2)
			self.assertEqual((yield dut.bus.cyc), 1)
			self.assertEqual((yield dut.bus.stb), 1)
			self.assertEqual((yield dut.bus.sel), 0b1100)
			self.assertEqual((yield dut.bus.we), 1)
			self.assertEqual((yield dut.bus.dat_w), 0x43218765)
			self.assertEqual((yield dut.bus.lock), 0)
			self.assertEqual((yield dut.bus.cti), CycleType.INCR_BURST.value)
			self.assertEqual((yield dut.bus.bte), BurstTypeExt.WRAP_4.value)
			self.assertEqual((yield intr_2.dat_r), 0xabcdef01)
			self.assertEqual((yield intr_2.ack), 1)
			self.assertEqual((yield intr_2.err), 1)
			self.assertEqual((yield intr_2.rty), 1)
			self.assertEqual((yield intr_2.stall), 0)

		sim = Simulator(dut)
		sim.add_clock(1e-6)
		sim.add_sync_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()

	def test_lock(self):
		dut = Arbiter(addr_width = 30, data_width = 32, features = {'lock'})
		intr_1 = Interface(addr_width = 30, data_width = 32, features = {'lock'})
		dut.add(intr_1)
		intr_2 = Interface(addr_width = 30, data_width = 32, features = {'lock'})
		dut.add(intr_2)

		def sim_test():
			yield intr_1.cyc.eq(1)
			yield intr_1.lock.eq(1)
			yield intr_2.cyc.eq(1)
			yield dut.bus.ack.eq(1)
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 1)
			self.assertEqual((yield intr_2.ack), 0)

			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 1)
			self.assertEqual((yield intr_2.ack), 0)

			yield intr_1.lock.eq(0)
			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 0)
			self.assertEqual((yield intr_2.ack), 1)

			yield intr_2.cyc.eq(0)
			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 1)
			self.assertEqual((yield intr_2.ack), 0)

			yield intr_1.stb.eq(1)
			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 1)
			self.assertEqual((yield intr_2.ack), 0)

			yield intr_1.stb.eq(0)
			yield intr_2.cyc.eq(1)
			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 0)
			self.assertEqual((yield intr_2.ack), 1)

		sim = Simulator(dut)
		sim.add_clock(1e-6)
		sim.add_sync_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()

	def test_stall(self):
		dut = Arbiter(addr_width = 30, data_width = 32, features = {'stall'})
		intr_1 = Interface(addr_width = 30, data_width = 32, features = {'stall'})
		dut.add(intr_1)
		intr_2 = Interface(addr_width = 30, data_width = 32, features = {'stall'})
		dut.add(intr_2)

		def sim_test():
			yield intr_1.cyc.eq(1)
			yield intr_2.cyc.eq(1)
			yield dut.bus.stall.eq(0)
			yield Delay(1e-6)
			self.assertEqual((yield intr_1.stall), 0)
			self.assertEqual((yield intr_2.stall), 1)

			yield dut.bus.stall.eq(1)
			yield Delay(1e-6)
			self.assertEqual((yield intr_1.stall), 1)
			self.assertEqual((yield intr_2.stall), 1)

		sim = Simulator(dut)
		sim.add_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()

	def test_stall_compat(self):
		dut = Arbiter(addr_width = 30, data_width = 32)
		intr_1 = Interface(addr_width = 30, data_width = 32, features = {'stall'})
		dut.add(intr_1)
		intr_2 = Interface(addr_width = 30, data_width = 32, features = {'stall'})
		dut.add(intr_2)

		def sim_test():
			yield intr_1.cyc.eq(1)
			yield intr_2.cyc.eq(1)
			yield Delay(1e-6)
			self.assertEqual((yield intr_1.stall), 1)
			self.assertEqual((yield intr_2.stall), 1)

			yield dut.bus.ack.eq(1)
			yield Delay(1e-6)
			self.assertEqual((yield intr_1.stall), 0)
			self.assertEqual((yield intr_2.stall), 1)

		sim = Simulator(dut)
		sim.add_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()

	def test_roundrobin(self):
		dut = Arbiter(addr_width = 30, data_width = 32)
		intr_1 = Interface(addr_width = 30, data_width = 32)
		dut.add(intr_1)
		intr_2 = Interface(addr_width = 30, data_width = 32)
		dut.add(intr_2)
		intr_3 = Interface(addr_width = 30, data_width = 32)
		dut.add(intr_3)

		def sim_test():
			yield intr_1.cyc.eq(1)
			yield intr_2.cyc.eq(0)
			yield intr_3.cyc.eq(1)
			yield dut.bus.ack.eq(1)
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 1)
			self.assertEqual((yield intr_2.ack), 0)
			self.assertEqual((yield intr_3.ack), 0)

			yield intr_1.cyc.eq(0)
			yield intr_2.cyc.eq(0)
			yield intr_3.cyc.eq(1)
			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 0)
			self.assertEqual((yield intr_2.ack), 0)
			self.assertEqual((yield intr_3.ack), 1)

			yield intr_1.cyc.eq(1)
			yield intr_2.cyc.eq(1)
			yield intr_3.cyc.eq(0)
			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 1)
			self.assertEqual((yield intr_2.ack), 0)
			self.assertEqual((yield intr_3.ack), 0)

			yield intr_1.cyc.eq(0)
			yield intr_2.cyc.eq(1)
			yield intr_3.cyc.eq(1)
			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 0)
			self.assertEqual((yield intr_2.ack), 1)
			self.assertEqual((yield intr_3.ack), 0)

			yield intr_1.cyc.eq(1)
			yield intr_2.cyc.eq(0)
			yield intr_3.cyc.eq(1)
			yield Tick()
			yield Delay(1e-7)
			self.assertEqual((yield intr_1.ack), 0)
			self.assertEqual((yield intr_2.ack), 0)
			self.assertEqual((yield intr_3.ack), 1)

		sim = Simulator(dut)
		sim.add_clock(1e-6)
		sim.add_sync_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()



# SPDX-License-Identifier: BSD-2-Clause

import re
import subprocess
import warnings
from pathlib import Path
from typing  import Callable, Optional

from .       import has_tool, require_tool

__all__ = (
	'find_yosys',
	'YosysBinary',
	'YosysError',
)

class YosysError(Exception):
	pass


class YosysWarning(Warning):
	pass


class YosysBinary:
	YOSYS_BINARY = 'yosys'

	@classmethod
	def available(cls) -> bool:
		'''
		Check for Yosys availability.

		Returns
		-------
		available : bool
			``True`` if Yosys is installed, ``False`` otherwise. Installed binary may still not
			be runnable, or might be too old to be useful.

		'''

		return has_tool(cls.YOSYS_BINARY)

	@classmethod
	def version(cls) -> Optional[tuple[int, int, int]]:
		'''
		Get Yosys version.

		Returns
		-------
		``None`` if version number could not be determined, or a 3-tuple ``(major, minor, distance)`` if it could.

		major : int
			Major version.
		minor : int
			Minor version.
		distance : int
			Distance to last tag per ``git describe``. May not be exact for system Yosys.

		'''

		version = cls.run(['-V'])
		match = re.match(r'^Yosys (\d+)\.(\d+)(?:\+(\d+))?', version)
		if match:
			return (int(match[1]), int(match[2]), int(match[3] or 0))
		else:
			return None

	@classmethod
	def data_dir(cls) -> Path:
		'''
		Get Yosys data directory.

		Returns
		-------
		data_dir : pathlib.Path
			Yosys data directory (also known as 'datdir').

		'''

		popen = subprocess.Popen(
			[require_tool(cls.YOSYS_BINARY) + '-config', '--datdir'],
			stdout = subprocess.PIPE, stderr = subprocess.PIPE,
			encoding = 'utf-8'
		)
		stdout, stderr = popen.communicate()
		if popen.returncode:
			raise YosysError(stderr.strip())
		return Path(stdout.strip())

	@classmethod
	def run(
		cls, args: list[str], stdin: str = '', *, ignore_warnings: bool = False, src_loc_at: int = 0
	) -> str:
		'''
		Run Yosys process.

		Parameters
		----------
		args : list of str
			Arguments, not including the program name.
		stdin : str
			Standard input.
		ignore_warnings : bool
			Ignore any warnings produce
		src_loc_at : int
			Source location

		Returns
		-------
		stdout : str
			Standard output.

		Exceptions
		----------
		YosysError
			Raised if Yosys returns a non-zero code. The exception message is the standard error
			output.

		'''

		popen = subprocess.Popen(
			[ require_tool(cls.YOSYS_BINARY), *args ],
			stdin = subprocess.PIPE, stdout = subprocess.PIPE, stderr = subprocess.PIPE,
			encoding = 'utf-8'
		)
		stdout, stderr = popen.communicate(stdin)
		# If Yosys is built with an evaluation version of Verific, then Verific license
		# information is printed first. It consists of empty lines and lines starting with `--`,
		# which are not normally a part of Yosys output, and can be fairly safely removed.
		#
		# This is not ideal, but Verific license conditions rule out any other solution.
		stdout = re.sub(r'\A(-- .+\n|\n)*', '', stdout)
		return cls._process_result(popen.returncode, stdout, stderr, ignore_warnings, src_loc_at)

	@classmethod
	def _process_result(
		cls, returncode: int, stdout: str, stderr: str, ignore_warnings: bool, src_loc_at: int
	) -> str:
		if returncode:
			raise YosysError(stderr.strip())
		if not ignore_warnings:
			for match in re.finditer(r'(?ms:^Warning: (.+)\n$)', stderr):
				message = match.group(1).replace('\n', ' ')
				warnings.warn(message, YosysWarning, stacklevel = 3 + src_loc_at)
		return stdout

def min_yosys_version(version: tuple[int, int, int]) -> bool:
	'''
	Returns if the yosys version is greater than or equal to the minimum
	required version

	Parameters
	----------
	version: tuple[int, int, int]
		The version of Yosys found on the system

	Returns
	-------
	bool:
		If the version meets the minimum requirement. (currently 0.15)

	'''

	return version >= (0, 15)

def find_yosys(
	requirement: Callable[[Optional[tuple[int, int, int]]], bool] = min_yosys_version
) -> YosysBinary:
	'''
	Find an available Yosys executable of required version.

	Parameters
	----------
	requirement : function
		Version check. Should return ``True`` if the version is acceptable, ``False`` otherwise.

	Returns
	-------
	yosys_binary : subclass of YosysBinary
		Proxy for running the requested version of Yosys.

	Exceptions
	----------
	YosysError
		Raised if required Yosys version is not found.

	'''


	if YosysBinary.available():
		version = YosysBinary.version()
		if version is not None and requirement(version):
			return YosysBinary
	raise YosysError(
		'Could not find an acceptable Yosys binary.\n'
		'Please ensure it is in your path or you set YOSYS environment variable.'
	)



# SPDX-License-Identifier: BSD-2-Clause

from inspect  import iscoroutinefunction, isgeneratorfunction
from typing   import IO, Coroutine, Generator, Iterable, Literal, Optional, Union
from warnings import warn

from ..hdl    import Signal
from ..hdl.cd import ClockDomain
from ..hdl.ir import Fragment
from ._base   import BaseEngine

__all__ = (
	'Active',
	'Delay',
	'Passive',
	'Settle',
	'Simulator',
	'Tick',
)


class Command:
	pass


class Settle(Command):
	'''
		micro-steps the simulation until just before the next clock cycle would start
	'''
	def __repr__(self) -> str:
		return '(settle)'


class Delay(Command):
	def __init__(self, interval: Optional[Union[int, float]] = None) -> None:
		self.interval = None if interval is None else float(interval)

	def __repr__(self) -> str:
		if self.interval is None:
			return '(delay ε)'
		else:
			return f'(delay {self.interval * 1e6:.3}us)'


class Tick(Command):
	'''
		Runs only the transition from high->low->high but does not progress execution
		further than that.

		.. warning::

			This can cause things to appear out-of-step.

	'''
	def __init__(self, domain: Union[str, ClockDomain] = 'sync') -> None:
		if not isinstance(domain, (str, ClockDomain)):
			raise TypeError(f'Domain must be a string or a ClockDomain instance, not {domain!r}')
		assert domain != 'comb'
		self.domain = domain

	def __repr__(self) -> str:
		return f'(tick {self.domain})'


class Passive(Command):
	def __repr__(self) -> str:
		return '(passive)'


class Active(Command):
	def __repr__(self) -> str:
		return '(active)'


class Simulator:
	def __init__(self, fragment: Fragment, *, engine: Literal['pysim'] = 'pysim') -> None:
		if isinstance(engine, type) and issubclass(engine, BaseEngine):
			pass
		elif engine == 'pysim':
			from .pysim import PySimEngine
			engine = PySimEngine
		else:
			raise TypeError(f'Value \'{engine!r}\' is not a simulation engine class or a simulation engine name')

		self._fragment = Fragment.get(fragment, platform = None).prepare()
		self._engine   = engine(self._fragment)
		self._clocked  = set()

	def _check_process(self, process: Union[Generator, Coroutine]) -> Union[Generator, Coroutine]:
		if not (isgeneratorfunction(process) or iscoroutinefunction(process)):
			raise TypeError(f'Cannot add a process {process!r} because it is not a generator function')
		return process

	def add_process(self, process: Union[Generator, Coroutine]) -> None:
		process = self._check_process(process)

		def wrapper():
			# Only start a bench process after comb settling, so that the reset values are correct.
			yield Settle()
			yield from process()

		self._engine.add_coroutine_process(wrapper, default_cmd = None)

	def add_sync_process(
		self, process: Union[Generator, Coroutine], *, domain: Union[str, ClockDomain] = 'sync'
	) -> None:
		process = self._check_process(process)

		def wrapper():
			# Only start a sync process after the first clock edge (or reset edge, if the domain
			# uses an asynchronous reset). This matches the behavior of synchronous FFs.
			yield Tick(domain)
			yield from process()

		self._engine.add_coroutine_process(wrapper, default_cmd = Tick(domain))

	def add_clock(
		self, period: float, *, phase: Optional[float] = None,
		domain: Union[str, ClockDomain] = 'sync', if_exists: bool = False
	) -> None:
		'''
		Add a clock process.

		Adds a process that drives the clock signal of ``domain`` at a 50% duty cycle.

		Arguments
		---------
		period : float
			Clock period. The process will toggle the ``domain`` clock signal every ``period / 2``
			seconds.
		phase : None or float
			Clock phase. The process will wait ``phase`` seconds before the first clock transition.
			If not specified, defaults to ``period / 2``.
		domain : str or ClockDomain
			Driven clock domain. If specified as a string, the domain with that name is looked up
			in the root fragment of the simulation.
		if_exists : bool
			If ``False`` (the default), raise an error if the driven domain is specified as
			a string and the root fragment does not have such a domain. If ``True``, do nothing
			in this case.

		'''

		if isinstance(domain, ClockDomain):
			if (domain.name in self._fragment.domains and
					domain is not self._fragment.domains[domain.name]):
				warn(
					'Adding a clock process that drives a clock domain object '
					f'named {domain.name!r}, which is distinct from an identically named domain '
					'in the simulated design',
					UserWarning, stacklevel = 2
				)
		elif domain in self._fragment.domains:
			domain = self._fragment.domains[domain]
		elif if_exists:
			return
		else:
			raise ValueError(f'Domain {domain!r} is not present in simulation')
		if domain in self._clocked:
			raise ValueError(f'Domain {domain.name!r} already has a clock driving it')

		# We represent times internally in 1 ps units, but users supply float quantities of seconds
		period = int(period * 1e12)

		if phase is None:
			# By default, delay the first edge by half period. This causes any synchronous activity
			# to happen at a non-zero time, distinguishing it from the reset values in the waveform
			# viewer.
			phase = period // 2
		else:
			phase = int(phase * 1e12) + period // 2
		self._engine.add_clock_process(domain.clk, phase = phase, period = period)
		self._clocked.add(domain)

	def reset(self) -> None:
		'''
		Reset the simulation.

		Assign the reset value to every signal in the simulation, and restart every user process.

		'''

		self._engine.reset()

	def advance(self) -> bool:
		'''
		Advance the simulation.

		Run every process and commit changes until a fixed point is reached, then advance time
		to the closest deadline (if any). If there is an unstable combinatorial loop,
		this function will never return.

		Returns ``True`` if there are any active processes, ``False`` otherwise.

		'''

		return self._engine.advance()

	def run(self) -> None:
		'''
		Run the simulation while any processes are active.

		Processes added with :meth:`add_process` and :meth:`add_sync_process` are initially active,
		and may change their status using the ``yield Passive()`` and ``yield Active()`` commands.
		Processes compiled from HDL and added with :meth:`add_clock` are always passive.

		'''

		while self.advance():
			pass

	def run_until(self, deadline: int, *, run_passive: bool = False) -> None:
		'''
		Run the simulation until it advances to ``deadline``.

		If ``run_passive`` is ``False``, the simulation also stops when there are no active
		processes, similar to :meth:`run`. Otherwise, the simulation will stop only after it
		advances to or past ``deadline``.

		If the simulation stops advancing, this function will never return.

		'''

		# Convert deadline in seconds into internal 1 ps units
		deadline = deadline * 1e12
		assert self._engine.now <= deadline
		while (self.advance() or run_passive) and self._engine.now < deadline:
			pass

	def write_vcd(
		self, vcd_file: Optional[Union[IO, str]], gtkw_file: Optional[Union[IO, str]] = None, *,
		traces: Iterable[Signal] = ()
	) -> None:
		'''
		Write waveforms to a Value Change Dump file, optionally populating a GTKWave save file.

		This method returns a context manager. It can be used as: ::

			sim = Simulator(frag)
			sim.add_clock(1e-6)
			with sim.write_vcd('dump.vcd", "dump.gtkw'):
				sim.run_until(1e-3)

		Arguments
		---------
		vcd_file : str or file-like object
			Verilog Value Change Dump file or filename.
		gtkw_file : str or file-like object
			GTKWave save file or filename.
		traces : iterable of Signal
			Signals to display traces for.

		'''

		if self._engine.now != 0:
			for file in (vcd_file, gtkw_file):
				if hasattr(file, 'close'):
					file.close()
			raise ValueError('Cannot start writing waveforms after advancing simulation time')

		return self._engine.write_vcd(vcd_file = vcd_file, gtkw_file = gtkw_file, traces = traces)



# SPDX-License-Identifier: BSD-2-Clause

from abc       import abstractmethod
from typing    import Literal, Union

from ...build  import Attrs, Clock, TemplatedPlatform
from ...hdl    import (
	ClockDomain, ClockSignal, Const, Instance, Module, Record, Signal
)
from ...lib.io import Pin

__all__ = (
	'IntelPlatform',
)

class IntelPlatform(TemplatedPlatform):
	'''
	.. rubric:: Quartus toolchain

	Required tools:
		* ``quartus_map``
		* ``quartus_fit``
		* ``quartus_asm``
		* ``quartus_sta``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_QUARTUS``, if present.

	Available overrides:
		* ``add_settings``: inserts commands at the end of the QSF file.
		* ``add_constraints``: inserts commands at the end of the SDC file.
		* ``nproc``: sets the number of cores used by all tools.
		* ``quartus_map_opts``: adds extra options for ``quartus_map``.
		* ``quartus_fit_opts``: adds extra options for ``quartus_fit``.
		* ``quartus_asm_opts``: adds extra options for ``quartus_asm``.
		* ``quartus_sta_opts``: adds extra options for ``quartus_sta``.

	Build products:
		* ``*.rpt``: toolchain reports.
		* ``{{name}}.sof``: bitstream as SRAM object file.
		* ``{{name}}.rbf``: bitstream as raw binary file.


	.. rubric:: Mistral toolchain

	Required tools:
		* ``yosys``
		* ``nextpnr-mistral``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_MISTRAL``, if present.

		* ``verbose``: enables logging of informational messages to standard error.
		* ``read_verilog_opts``: adds options for ``read_verilog`` Yosys command.
		* ``synth_opts``: adds options for ``synth_intel_alm`` Yosys command.
		* ``script_after_read``: inserts commands after ``read_rtlil`` in Yosys script.
		* ``script_after_synth``: inserts commands after ``synth_intel_alm`` in Yosys script.
		* ``yosys_opts``: adds extra options for ``yosys``.
		* ``nextpnr_opts``: adds extra options for ``nextpnr-mistral``.

	'''

	toolchain = None # selected when creating platform

	@property
	@abstractmethod
	def device(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def package(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def speed(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	suffix  = ''

	# Quartus templates

	quartus_suppressed_warnings = [
		10264,  # All case item expressions in this case statement are onehot
		10270,  # Incomplete Verilog case statement has no default case item
		10335,  # Unrecognized synthesis attribute
		10763,  # Verilog case statement has overlapping case item expressions with non-constant or don't care bits
		10935,  # Verilog casex/casez overlaps with a previous casex/vasez item expression
		12125,  # Using design file which is not specified as a design file for the current project, but contains
				# definitions used in project
		18236,  # Number of processors not specified in QSF
		292013, # Feature is only available with a valid subscription license
	]

	quartus_required_tools = [
		'quartus_map',
		'quartus_fit',
		'quartus_asm',
		'quartus_sta',
	]

	quartus_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'build_{{name}}.sh': r'''
			# {{autogenerated}}
			if [ -n "${{platform._toolchain_env_var}}" ]; then
				QUARTUS_ROOTDIR=$(dirname $(dirname "${{platform._toolchain_env_var}}"))
				# Quartus' qenv.sh does not work with `set -e`.
				. "${{platform._toolchain_env_var}}"
			fi
			set -e{{verbose("x")}}
			{{emit_commands("sh")}}
		''',
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.qsf': r'''
			# {{autogenerated}}
			{% if get_override("nproc") -%}
				set_global_assignment -name NUM_PARALLEL_PROCESSORS {{get_override("nproc")}}
			{% endif %}

			{% for file in platform.iter_files(".v") -%}
				set_global_assignment -name VERILOG_FILE {{file|tcl_quote}}
			{% endfor %}
			{% for file in platform.iter_files(".sv") -%}
				set_global_assignment -name SYSTEMVERILOG_FILE {{file|tcl_quote}}
			{% endfor %}
			{% for file in platform.iter_files(".vhd", ".vhdl") -%}
				set_global_assignment -name VHDL_FILE {{file|tcl_quote}}
			{% endfor %}
			set_global_assignment -name VERILOG_FILE {{name}}.v
			set_global_assignment -name TOP_LEVEL_ENTITY {{name}}

			set_global_assignment -name DEVICE {{platform.device}}{{platform.package}}{{platform.speed}}{{platform.suffix}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				set_location_assignment -to {{port_name|tcl_quote}} PIN_{{pin_name}}
				{% for key, value in attrs.items() -%}
					set_instance_assignment -to {{port_name|tcl_quote}} -name {{key}} {{value|tcl_quote}}
				{% endfor %}
			{% endfor %}

			set_global_assignment -name GENERATE_RBF_FILE ON

			{{get_override("add_settings")|default("# (add_settings placeholder)")}}
		''',
		'{{name}}.sdc': r'''
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				{% if port_signal is not none -%}
					create_clock -name {{port_signal.name|tcl_quote}} -period {{1000000000/frequency}} [get_ports {{port_signal.name|tcl_quote}}]
				{% else -%}
					create_clock -name {{net_signal.name|tcl_quote}} -period {{1000000000/frequency}} [get_nets {{net_signal|hierarchy("|")|tcl_quote}}]
				{% endif %}
			{% endfor %}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''', # noqa: E501
		'{{name}}.srf': r'''
			{% for warning in platform.quartus_suppressed_warnings %}
			{ "" "" "" "{{name}}.v" {  } {  } 0 {{warning}} "" 0 0 "Design Software" 0 -1 0 ""}
			{% endfor %}
		''',
	}
	quartus_command_templates = [
		r'''
		{{invoke_tool("quartus_map")}}
			{{get_override("quartus_map_opts")|options}}
			--rev={{name}} {{name}}
		''',
		r'''
		{{invoke_tool("quartus_fit")}}
			{{get_override("quartus_fit_opts")|options}}
			--rev={{name}} {{name}}
		''',
		r'''
		{{invoke_tool("quartus_asm")}}
			{{get_override("quartus_asm_opts")|options}}
			--rev={{name}} {{name}}
		''',
		r'''
		{{invoke_tool("quartus_sta")}}
			{{get_override("quartus_sta_opts")|options}}
			--rev={{name}} {{name}}
		''',
	]


	# Mistral templates

	mistral_required_tools = [
		'yosys',
		'nextpnr-mistral'
	]
	mistral_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'{{name}}.il': r'''
			# {{autogenerated}}
			{{emit_rtlil()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.ys': r'''
			# {{autogenerated}}
			{% for file in platform.iter_files(".v") -%}
				read_verilog {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".sv") -%}
				read_verilog -sv {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".il") -%}
				read_rtlil {{file}}
			{% endfor %}
			read_rtlil {{name}}.il
			delete w:$verilog_initial_trigger
			{{get_override("script_after_read")|default("# (script_after_read placeholder)")}}
			synth_intel_alm {{get_override("synth_opts")|options}} -top {{name}}
			{{get_override("script_after_synth")|default("# (script_after_synth placeholder)")}}
			write_json {{name}}.json
		''',
		'{{name}}.qsf': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				set_location_assignment -to {{port_name|tcl_quote}} PIN_{{pin_name}}
				{% for key, value in attrs.items() -%}
					set_instance_assignment -to {{port_name|tcl_quote}} -name {{key}} {{value|tcl_quote}}
				{% endfor %}
			{% endfor %}
		''',

	}
	mistral_command_templates = [
		r'''
		{{invoke_tool("yosys")}}
			{{quiet("-q")}}
			{{get_override("yosys_opts")|options}}
			-l {{name}}.rpt
			{{name}}.ys
		''',
		r'''
		{{invoke_tool("nextpnr-mistral")}}
			{{quiet("--quiet")}}
			{{get_override("nextpnr_opts")|options}}
			--log {{name}}.tim
			--device {{platform.device}}{{platform.package}}{{platform.speed}}{{platform.suffix}}
			--json {{name}}.json
			--qsf {{name}}.qsf
			--rbf {{name}}.rbf
		'''
	]

	# Common logic

	def __init__(
		self, *, toolchain: Literal['Quartus', 'Mistral'] = 'Quartus'
	) -> None:
		super().__init__()

		assert toolchain in ('Quartus', 'Mistral')
		self.toolchain = toolchain

	@property
	def required_tools(self) -> list[str]:
		if self.toolchain == 'Quartus':
			return self.quartus_required_tools
		if self.toolchain == 'Mistral':
			return self.mistral_required_tools
		assert False

	@property
	def file_templates(self) -> dict[str, str]:
		if self.toolchain == 'Quartus':
			return self.quartus_file_templates
		if self.toolchain == 'Mistral':
			return self.mistral_file_templates
		assert False

	@property
	def command_templates(self) -> list[str]:
		if self.toolchain == 'Quartus':
			return self.quartus_command_templates
		if self.toolchain == 'Mistral':
			return self.mistral_command_templates
		assert False

	def add_clock_constraint(self, clock: Signal, frequency: Union[int, float]) -> None:
		super().add_clock_constraint(clock, frequency)
		clock.attrs['keep'] = 'true'

	@property
	def default_clk_constraint(self) -> Clock:
		# Internal high-speed oscillator on Cyclone V devices.
		# It is specified to not be faster than 100MHz, but the actual
		# frequency seems to vary a lot between devices. Measurements
		# of 78 to 84 MHz have been observed.
		if self.default_clk == 'cyclonev_oscillator':
			assert self.device.startswith('5C')
			return Clock(100e6)
		# Otherwise, use the defined Clock resource.
		return super().default_clk_constraint

	def create_missing_domain(self, name: str) -> Module:
		if name == 'sync' and self.default_clk == 'cyclonev_oscillator':
			# Use the internal high-speed oscillator for Cyclone V devices
			assert self.device.startswith('5C')
			m = Module()
			m.domains += ClockDomain('sync')
			m.submodules += Instance(
				'cyclonev_oscillator',
				i_oscena = Const(1),
				o_clkout = ClockSignal('sync')
			)
			return m
		else:
			return super().create_missing_domain(name)

	# The altiobuf_* and altddio_* primitives are explained in the following Intel documents:
	# * https://www.intel.com/content/dam/www/programmable/us/en/pdfs/literature/ug/ug_altiobuf.pdf
	# * https://www.intel.com/content/dam/www/programmable/us/en/pdfs/literature/ug/ug_altddio.pdf
	# See also errata mentioned in:
	# 	https://www.intel.com/content/www/us/en/programmable/support/support-resources/knowledge-base/solutions/rd11192012_735.html.

	@staticmethod
	def _get_ireg(m: Module, pin: Pin, invert: bool) -> Signal:
		def get_ineg(i: Signal):
			if invert:
				i_neg = Signal.like(i, name_suffix = '_neg')
				m.d.comb += i.eq(~i_neg)
				return i_neg
			else:
				return i

		if pin.xdr == 0:
			return get_ineg(pin.i)
		elif pin.xdr == 1:
			i_sdr = Signal(pin.width, name = f'{pin.name}_i_sdr')
			m.submodules += Instance(
				'$dff',
				p_CLK_POLARITY = 1,
				p_WIDTH = pin.width,
				i_CLK = pin.i_clk,
				i_D = i_sdr,
				o_Q = get_ineg(pin.i),
			)
			return i_sdr
		elif pin.xdr == 2:
			i_ddr = Signal(pin.width, name = f'{pin.name}_i_ddr')
			m.submodules[f'{pin.name}_i_ddr'] = Instance(
				'altddio_in',
				p_width = pin.width,
				i_datain = i_ddr,
				i_inclock = pin.i_clk,
				o_dataout_h = get_ineg(pin.i0),
				o_dataout_l = get_ineg(pin.i1),
			)
			return i_ddr
		assert False

	@staticmethod
	def _get_oreg(m: Module, pin: Pin, invert: bool) -> Signal:
		def get_oneg(o: Signal):
			if invert:
				o_neg = Signal.like(o, name_suffix = '_neg')
				m.d.comb += o_neg.eq(~o)
				return o_neg
			else:
				return o

		if pin.xdr == 0:
			return get_oneg(pin.o)
		elif pin.xdr == 1:
			o_sdr = Signal(pin.width, name = f'{pin.name}_o_sdr')
			m.submodules += Instance(
				'$dff',
				p_CLK_POLARITY = 1,
				p_WIDTH = pin.width,
				i_CLK = pin.o_clk,
				i_D = get_oneg(pin.o),
				o_Q = o_sdr,
			)
			return o_sdr
		elif pin.xdr == 2:
			o_ddr = Signal(pin.width, name = f'{pin.name}_o_ddr')
			m.submodules[f'{pin.name}_o_ddr'] = Instance(
				'altddio_out',
				p_width = pin.width,
				o_dataout = o_ddr,
				i_outclock = pin.o_clk,
				i_datain_h = get_oneg(pin.o0),
				i_datain_l = get_oneg(pin.o1),
			)
			return o_ddr
		assert False

	@staticmethod
	def _get_oereg(m: Module, pin: Pin) -> Signal:
		# altiobuf_ requires an output enable signal for each pin, but pin.oe is 1 bit wide.
		if pin.xdr == 0:
			return pin.oe.replicate(pin.width)
		elif pin.xdr in (1, 2):
			oe_reg = Signal(pin.width, name = f'{pin.name}_oe_reg')
			oe_reg.attrs['useioff'] = '1'
			m.submodules += Instance(
				'$dff',
				p_CLK_POLARITY = 1,
				p_WIDTH = pin.width,
				i_CLK = pin.o_clk,
				i_D = pin.oe,
				o_Q = oe_reg,
			)
			return oe_reg
		assert False

	def get_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		if pin.xdr == 1:
			port.attrs['useioff'] = 1

		m = Module()
		m.submodules[pin.name] = Instance(
			'altiobuf_in',
			p_enable_bus_hold = 'FALSE',
			p_number_of_channels = pin.width,
			p_use_differential_mode = 'FALSE',
			i_datain = port.io,
			o_dataout = self._get_ireg(m, pin, invert)
		)
		return m

	def get_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		if pin.xdr == 1:
			port.attrs['useioff'] = 1

		m = Module()
		m.submodules[pin.name] = Instance(
			'altiobuf_out',
			p_enable_bus_hold = 'FALSE',
			p_number_of_channels = pin.width,
			p_use_differential_mode = 'FALSE',
			p_use_oe = 'FALSE',
			i_datain = self._get_oreg(m, pin, invert),
			o_dataout = port.io,
		)
		return m

	def get_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended tristate', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		if pin.xdr == 1:
			port.attrs['useioff'] = 1

		m = Module()
		m.submodules[pin.name] = Instance(
			'altiobuf_out',
			p_enable_bus_hold = 'FALSE',
			p_number_of_channels = pin.width,
			p_use_differential_mode = "FALSE",
			p_use_oe = 'TRUE',
			i_datain = self._get_oreg(m, pin, invert),
			o_dataout = port.io,
			i_oe = self._get_oereg(m, pin)
		)
		return m

	def get_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input/output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		if pin.xdr == 1:
			port.attrs['useioff'] = 1

		m = Module()
		m.submodules[pin.name] = Instance(
			'altiobuf_bidir',
			p_enable_bus_hold = 'FALSE',
			p_number_of_channels = pin.width,
			p_use_differential_mode = 'FALSE',
			i_datain = self._get_oreg(m, pin, invert),
			io_dataio = port.io,
			o_dataout = self._get_ireg(m, pin, invert),
			i_oe = self._get_oereg(m, pin),
		)
		return m

	def get_diff_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential input', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		if pin.xdr == 1:
			port.p.attrs['useioff'] = 1
			port.n.attrs['useioff'] = 1

		m = Module()
		m.submodules[pin.name] = Instance(
			'altiobuf_in',
			p_enable_bus_hold = 'FALSE',
			p_number_of_channels = pin.width,
			p_use_differential_mode = 'TRUE',
			i_datain = port.p,
			i_datain_b = port.n,
			o_dataout = self._get_ireg(m, pin, invert)
		)
		return m

	def get_diff_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		if pin.xdr == 1:
			port.p.attrs['useioff'] = 1
			port.n.attrs['useioff'] = 1

		m = Module()
		m.submodules[pin.name] = Instance('altiobuf_out',
			p_enable_bus_hold = 'FALSE',
			p_number_of_channels = pin.width,
			p_use_differential_mode = 'TRUE',
			p_use_oe = 'FALSE',
			i_datain = self._get_oreg(m, pin, invert),
			o_dataout = port.p,
			o_dataout_b = port.n,
		)
		return m

	def get_diff_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential tristate', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		if pin.xdr == 1:
			port.p.attrs['useioff'] = 1
			port.n.attrs['useioff'] = 1

		m = Module()
		m.submodules[pin.name] = Instance(
			'altiobuf_out',
			p_enable_bus_hold = 'FALSE',
			p_number_of_channels = pin.width,
			p_use_differential_mode = 'TRUE',
			p_use_oe = 'TRUE',
			i_datain = self._get_oreg(m, pin, invert),
			o_dataout = port.p,
			o_dataout_b = port.n,
			i_oe = self._get_oereg(m, pin),
		)
		return m

	def get_diff_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential input/output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		if pin.xdr == 1:
			port.p.attrs['useioff'] = 1
			port.n.attrs['useioff'] = 1

		m = Module()
		m.submodules[pin.name] = Instance(
			'altiobuf_bidir',
			p_enable_bus_hold = 'FALSE',
			p_number_of_channels = pin.width,
			p_use_differential_mode = 'TRUE',
			i_datain = self._get_oreg(m, pin, invert),
			io_dataio = port.p,
			io_dataio_b = port.n,
			o_dataout = self._get_ireg(m, pin, invert),
			i_oe = self._get_oereg(m, pin),
		)
		return m

	# The altera_std_synchronizer{,_bundle} megafunctions embed SDC constraints that mark false
	# paths, so use them instead of our default implementation.

	def get_ff_sync(self, ff_sync) -> Instance:
		return Instance(
			'altera_std_synchronizer_bundle',
			p_width = len(ff_sync.i),
			p_depth = ff_sync._stages,
			i_clk = ClockSignal(ff_sync._o_domain),
			i_reset_n = Const(1),
			i_din = ff_sync.i,
			o_dout = ff_sync.o,
		)

	def get_async_ff_sync(self, async_ff_sync) -> Module:
		m = Module()
		sync_output = Signal()
		if async_ff_sync._edge == 'pos':
			m.submodules += Instance(
				'altera_std_synchronizer',
				p_depth = async_ff_sync._stages,
				i_clk = ClockSignal(async_ff_sync._o_domain),
				i_reset_n = ~async_ff_sync.i,
				i_din = Const(1),
				o_dout = sync_output,
			)
		else:
			m.submodules += Instance(
				'altera_std_synchronizer',
				p_depth = async_ff_sync._stages,
				i_clk = ClockSignal(async_ff_sync._o_domain),
				i_reset_n = async_ff_sync.i,
				i_din = Const(1),
				o_dout = sync_output,
			)
		m.d.comb += async_ff_sync.o.eq(~sync_output)
		return m



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii                      import *
from torii.sim                  import *
from torii.lib.soc              import csr
from torii.lib.soc.csr.wishbone import *

from ....utils                  import ToriiTestSuiteCase
class MockRegister(Elaboratable):
	def __init__(self, width, name):
		self.element = csr.Element(width, 'rw', name = name)
		self.r_count = Signal(8)
		self.w_count = Signal(8)
		self.data    = Signal(width)

	def elaborate(self, platform):
		m = Module()

		with m.If(self.element.r_stb):
			m.d.sync += self.r_count.eq(self.r_count + 1)
		m.d.comb += self.element.r_data.eq(self.data)

		with m.If(self.element.w_stb):
			m.d.sync += self.w_count.eq(self.w_count + 1)
			m.d.sync += self.data.eq(self.element.w_data)

		return m


class WishboneCSRBridgeTestCase(ToriiTestSuiteCase):
	def test_wrong_csr_bus(self):
		with self.assertRaisesRegex(
			ValueError,
			r'CSR bus must be an instance of CSRInterface, not \'foo\''
		):
			WishboneCSRBridge('foo')

	def test_wrong_csr_bus_data_width(self):
		with self.assertRaisesRegex(
			ValueError,
			r'CSR bus data width must be one of 8, 16, 32, 64, not 7'
		):
			WishboneCSRBridge(csr_bus = csr.Interface(addr_width = 10, data_width = 7))

	def test_narrow(self):
		mux   = csr.Multiplexer(addr_width = 10, data_width = 8)
		reg_1 = MockRegister(8, name = 'reg_1')
		mux.add(reg_1.element)
		reg_2 = MockRegister(16, name = 'reg_2')
		mux.add(reg_2.element)
		dut   = WishboneCSRBridge(mux.bus)

		def sim_test():
			yield dut.wb_bus.cyc.eq(1)
			yield dut.wb_bus.sel.eq(0b1)

			yield dut.wb_bus.we.eq(1)

			yield dut.wb_bus.adr.eq(0)
			yield dut.wb_bus.stb.eq(1)
			yield dut.wb_bus.dat_w.eq(0x55)
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg_1.r_count), 0)
			self.assertEqual((yield reg_1.w_count), 1)
			self.assertEqual((yield reg_1.data), 0x55)

			yield dut.wb_bus.adr.eq(1)
			yield dut.wb_bus.stb.eq(1)
			yield dut.wb_bus.dat_w.eq(0xaa)
			yield
			yield
			yield
			yield dut.wb_bus.stb.eq(0)
			self.assertEqual((yield dut.wb_bus.ack), 1)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg_2.r_count), 0)
			self.assertEqual((yield reg_2.w_count), 0)
			self.assertEqual((yield reg_2.data), 0)

			yield dut.wb_bus.adr.eq(2)
			yield dut.wb_bus.stb.eq(1)
			yield dut.wb_bus.dat_w.eq(0xbb)
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg_2.r_count), 0)
			self.assertEqual((yield reg_2.w_count), 1)
			self.assertEqual((yield reg_2.data), 0xbbaa)

			yield dut.wb_bus.we.eq(0)

			yield dut.wb_bus.adr.eq(0)
			yield dut.wb_bus.stb.eq(1)
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			self.assertEqual((yield dut.wb_bus.dat_r), 0x55)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg_1.r_count), 1)
			self.assertEqual((yield reg_1.w_count), 1)

			yield dut.wb_bus.adr.eq(1)
			yield dut.wb_bus.stb.eq(1)
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			self.assertEqual((yield dut.wb_bus.dat_r), 0xaa)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg_2.r_count), 1)
			self.assertEqual((yield reg_2.w_count), 1)

			yield reg_2.data.eq(0x33333)

			yield dut.wb_bus.adr.eq(2)
			yield dut.wb_bus.stb.eq(1)
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			self.assertEqual((yield dut.wb_bus.dat_r), 0xbb)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg_2.r_count), 1)
			self.assertEqual((yield reg_2.w_count), 1)

		m = Module()
		m.submodules += mux, reg_1, reg_2, dut
		sim = Simulator(m)
		sim.add_clock(1e-6)
		sim.add_sync_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()

	def test_wide(self):
		mux = csr.Multiplexer(addr_width = 10, data_width = 8)
		reg = MockRegister(32, name = 'reg')
		mux.add(reg.element)
		dut = WishboneCSRBridge(mux.bus, data_width = 32)

		def sim_test():
			yield dut.wb_bus.cyc.eq(1)
			yield dut.wb_bus.adr.eq(0)

			yield dut.wb_bus.we.eq(1)

			yield dut.wb_bus.dat_w.eq(0x44332211)
			yield dut.wb_bus.sel.eq(0b1111)
			yield dut.wb_bus.stb.eq(1)
			yield
			yield
			yield
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg.r_count), 0)
			self.assertEqual((yield reg.w_count), 1)
			self.assertEqual((yield reg.data), 0x44332211)

			# partial write
			yield dut.wb_bus.dat_w.eq(0xaabbccdd)
			yield dut.wb_bus.sel.eq(0b0110)
			yield dut.wb_bus.stb.eq(1)
			yield
			yield
			yield
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg.r_count), 0)
			self.assertEqual((yield reg.w_count), 1)
			self.assertEqual((yield reg.data), 0x44332211)

			yield dut.wb_bus.we.eq(0)

			yield dut.wb_bus.sel.eq(0b1111)
			yield dut.wb_bus.stb.eq(1)
			yield
			yield
			yield
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			self.assertEqual((yield dut.wb_bus.dat_r), 0x44332211)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg.r_count), 1)
			self.assertEqual((yield reg.w_count), 1)

			yield reg.data.eq(0xaaaaaaaa)

			# partial read
			yield dut.wb_bus.sel.eq(0b0110)
			yield dut.wb_bus.stb.eq(1)
			yield
			yield
			yield
			yield
			yield
			yield
			self.assertEqual((yield dut.wb_bus.ack), 1)
			self.assertEqual((yield dut.wb_bus.dat_r), 0x00332200)
			yield dut.wb_bus.stb.eq(0)
			yield
			self.assertEqual((yield dut.wb_bus.ack), 0)
			self.assertEqual((yield reg.r_count), 1)
			self.assertEqual((yield reg.w_count), 1)

		m = Module()
		m.submodules += mux, reg, dut
		sim = Simulator(m)
		sim.add_clock(1e-6)
		sim.add_sync_process(sim_test)
		with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
			sim.run()



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii.hdl     import *
from torii.sim     import *
from torii.lib.cdc import *

from ..utils       import ToriiTestSuiteCase

class FFSynchronizerTestCase(ToriiTestSuiteCase):
	def test_stages_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Synchronization stage count must be a positive integer, not 0$'
		):
			FFSynchronizer(Signal(), Signal(), stages = 0)
		with self.assertRaisesRegex(
			ValueError,
			r'^Synchronization stage count may not safely be less than 2$'
		):
			FFSynchronizer(Signal(), Signal(), stages = 1)

	def test_basic(self):
		i = Signal()
		o = Signal()
		frag = FFSynchronizer(i, o)

		sim = Simulator(frag)
		sim.add_clock(1e-6)

		def process():
			self.assertEqual((yield o), 0)
			yield i.eq(1)
			yield Tick()
			self.assertEqual((yield o), 0)
			yield Tick()
			self.assertEqual((yield o), 0)
			yield Tick()
			self.assertEqual((yield o), 1)
		sim.add_process(process)
		sim.run()

	def test_reset_value(self):
		i = Signal(reset = 1)
		o = Signal()
		frag = FFSynchronizer(i, o, reset = 1)

		sim = Simulator(frag)
		sim.add_clock(1e-6)

		def process():
			self.assertEqual((yield o), 1)
			yield i.eq(0)
			yield Tick()
			self.assertEqual((yield o), 1)
			yield Tick()
			self.assertEqual((yield o), 1)
			yield Tick()
			self.assertEqual((yield o), 0)
		sim.add_process(process)
		sim.run()


class AsyncFFSynchronizerTestCase(ToriiTestSuiteCase):
	def test_stages_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Synchronization stage count must be a positive integer, not 0$'
		):
			ResetSynchronizer(Signal(), stages = 0)
		with self.assertRaisesRegex(
			ValueError,
			r'^Synchronization stage count may not safely be less than 2$'
		):
			ResetSynchronizer(Signal(), stages = 1)

	def test_edge_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^AsyncFFSynchronizer async edge must be one of \'pos\' or \'neg\', not \'xxx\'$'
		):
			AsyncFFSynchronizer(Signal(), Signal(), o_domain = 'sync', async_edge = 'xxx')

	def test_width_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^AsyncFFSynchronizer input width must be 1, not 2$'
		):
			AsyncFFSynchronizer(Signal(2), Signal(), o_domain = 'sync')
		with self.assertRaisesRegex(
			ValueError,
			r'^AsyncFFSynchronizer output width must be 1, not 2$'
		):
			AsyncFFSynchronizer(Signal(), Signal(2), o_domain = 'sync')

	def test_pos_edge(self):
		i = Signal()
		o = Signal()
		m = Module()
		m.domains += ClockDomain('sync')
		m.submodules += AsyncFFSynchronizer(i, o)

		sim = Simulator(m)
		sim.add_clock(1e-6)

		def process():
			# initial reset
			self.assertEqual((yield i), 0)
			self.assertEqual((yield o), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 0)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 0)
			yield Tick()
			yield Delay(1e-8)

			yield i.eq(1)
			yield Delay(1e-8)
			self.assertEqual((yield o), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 1)
			yield i.eq(0)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 0)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 0)
			yield Tick()
			yield Delay(1e-8)
		sim.add_process(process)
		with sim.write_vcd('test.vcd'):
			sim.run()

	def test_neg_edge(self):
		i = Signal(reset = 1)
		o = Signal()
		m = Module()
		m.domains += ClockDomain('sync')
		m.submodules += AsyncFFSynchronizer(i, o, async_edge = 'neg')

		sim = Simulator(m)
		sim.add_clock(1e-6)

		def process():
			# initial reset
			self.assertEqual((yield i), 1)
			self.assertEqual((yield o), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 0)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 0)
			yield Tick()
			yield Delay(1e-8)

			yield i.eq(0)
			yield Delay(1e-8)
			self.assertEqual((yield o), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 1)
			yield i.eq(1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 0)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield o), 0)
			yield Tick()
			yield Delay(1e-8)
		sim.add_process(process)
		with sim.write_vcd('test.vcd'):
			sim.run()


class ResetSynchronizerTestCase(ToriiTestSuiteCase):
	def test_stages_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Synchronization stage count must be a positive integer, not 0$'
		):
			ResetSynchronizer(Signal(), stages = 0)
		with self.assertRaisesRegex(
			ValueError,
			r'^Synchronization stage count may not safely be less than 2$'
		):
			ResetSynchronizer(Signal(), stages = 1)

	def test_basic(self):
		arst = Signal()
		m = Module()
		m.domains += ClockDomain('sync')
		m.submodules += ResetSynchronizer(arst)
		s = Signal(reset = 1)
		m.d.sync += s.eq(0)

		sim = Simulator(m)
		sim.add_clock(1e-6)

		def process():
			# initial reset
			self.assertEqual((yield s), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield s), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield s), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield s), 0)
			yield Tick()
			yield Delay(1e-8)

			yield arst.eq(1)
			yield Delay(1e-8)
			self.assertEqual((yield s), 0)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield s), 1)
			yield arst.eq(0)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield s), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield s), 1)
			yield Tick()
			yield Delay(1e-8)
			self.assertEqual((yield s), 0)
			yield Tick()
			yield Delay(1e-8)
		sim.add_process(process)
		with sim.write_vcd('test.vcd'):
			sim.run()


# TODO: test with distinct clocks
class PulseSynchronizerTestCase(ToriiTestSuiteCase):
	def test_stages_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Synchronization stage count must be a positive integer, not 0$'
		):
			PulseSynchronizer('w', 'r', stages = 0)
		with self.assertRaisesRegex(
			ValueError,
			r'^Synchronization stage count may not safely be less than 2$'
		):
			PulseSynchronizer('w', 'r', stages = 1)

	def test_smoke(self):
		m = Module()
		m.domains += ClockDomain('sync')
		ps = m.submodules.dut = PulseSynchronizer('sync', 'sync')

		sim = Simulator(m)
		sim.add_clock(1e-6)

		def process():
			yield ps.i.eq(0)
			# TODO: think about reset
			for n in range(5):
				yield Tick()
			# Make sure no pulses are generated in quiescent state
			for n in range(3):
				yield Tick()
				self.assertEqual((yield ps.o), 0)
			# Check conservation of pulses
			accum = 0
			for n in range(10):
				yield ps.i.eq(1 if n < 4 else 0)
				yield Tick()
				accum += yield ps.o
			self.assertEqual(accum, 4)
		sim.add_process(process)
		sim.run()



# SPDX-License-Identifier: BSD-2-Clause

from ...      import Elaboratable, Module, Signal
from .one_hot import Decoder as OneHotDecoder

__all__ = (
	'Decoder',
	'Encoder',
)


class Encoder(Elaboratable):
	'''
	Priority encode requests to binary.

	If any bit in ``i`` is asserted, ``n`` is low and ``o`` indicates the least significant
	asserted bit.
	Otherwise, ``n`` is high and ``o`` is ``0``.

	Parameters
	----------
	width : int
		Bit width of the input.

	Attributes
	----------
	i : Signal(width), in
		Input requests.
	o : Signal(range(width)), out
		Encoded natural binary.
	n : Signal, out
		Invalid: no input bits are asserted.

	'''

	def __init__(self, width: int) -> None:
		self.width = width

		self.i = Signal(width)
		self.o = Signal(range(width))
		self.n = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()
		for j in reversed(range(self.width)):
			with m.If(self.i[j]):
				m.d.comb += self.o.eq(j)
		m.d.comb += self.n.eq(self.i == 0)
		return m


class Decoder(OneHotDecoder):
	'''
	Decode binary to priority request.

	Identical to :py:class:`torii.lib.coding.one_hot.Decoder`.

	'''



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Signal, Module
from torii.back import verilog


class Counter(Elaboratable):
	def __init__(self, width: int):
		self.v = Signal(width, reset = 2**width-1)
		self.o = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()
		m.d.sync += self.v.eq(self.v + 1)
		m.d.comb += self.o.eq(self.v[-1])
		return m


ctr = Counter(width = 16)
if __name__ == '__main__':
	print(verilog.convert(ctr, ports = [ctr.o]))



# SPDX-License-Identifier: BSD-2-Clause

from contextlib import contextmanager
from os         import getenv
from sys        import version_info
from tempfile   import NamedTemporaryFile

from ..hdl.ast  import SignalSet
from ..hdl.xfrm import LHSGroupFilter, StatementVisitor, ValueVisitor
from ._base     import BaseProcess

__all__ = (
	'PyRTLProcess',
)

_USE_PATTERN_MATCHING = (version_info >= (3, 10))

class PyRTLProcess(BaseProcess):
	__slots__ = ('is_comb', 'runnable', 'passive', 'run')

	def __init__(self, *, is_comb):
		self.is_comb  = is_comb

		self.reset()

	def reset(self):
		self.runnable = self.is_comb
		self.passive  = True


class _PythonEmitter:
	def __init__(self):
		self._buffer = []
		self._suffix = 0
		self._level  = 0

	def append(self, code):
		self._buffer.append('\t' * self._level)
		self._buffer.append(code)
		self._buffer.append('\n')

	@contextmanager
	def indent(self):
		self._level += 1
		yield
		self._level -= 1

	def flush(self, indent = ''):
		code = ''.join(self._buffer)
		self._buffer.clear()
		return code

	def gen_var(self, prefix):
		name = f'{prefix}_{self._suffix}'
		self._suffix += 1
		return name

	def def_var(self, prefix, value):
		name = self.gen_var(prefix)
		self.append(f'{name} = {value}')
		return name


class _Compiler:
	def __init__(self, state, emitter):
		self.state = state
		self.emitter = emitter


class _ValueCompiler(ValueVisitor, _Compiler):
	helpers = {
		'sign': lambda value, sign: value | sign if value & sign else value,
		'zdiv': lambda lhs, rhs: 0 if rhs == 0 else lhs // rhs,
		'zmod': lambda lhs, rhs: 0 if rhs == 0 else lhs % rhs,
	}

	def on_value(self, value):
		# Very large values are unlikely to compile or simulate in reasonable time.
		if len(value) > 2 ** 16:
			if value.src_loc:
				src = '{}:{}'.format(*value.src_loc)
			else:
				src = 'unknown location'
			raise OverflowError(
				f'Value defined at {src} is {len(value)} bits wide, which is unlikely to simulate '
				'in reasonable time'
			)

		val = super().on_value(value)
		if isinstance(val, str) and len(val) > 1000:
			# Avoid parser stack overflow on older Pythons.
			return self.emitter.def_var('intermediate', val)
		return val

	def on_ClockSignal(self, value):
		raise NotImplementedError # :nocov:

	def on_ResetSignal(self, value):
		raise NotImplementedError # :nocov:

	def on_AnyConst(self, value):
		raise NotImplementedError # :nocov:

	def on_AnySeq(self, value):
		raise NotImplementedError # :nocov:

	def on_Sample(self, value):
		raise NotImplementedError # :nocov:

	def on_Initial(self, value):
		raise NotImplementedError # :nocov:


class _RHSValueCompiler(_ValueCompiler):
	def __init__(self, state, emitter, *, mode, inputs = None):
		super().__init__(state, emitter)
		assert mode in ('curr', 'next')
		self.mode = mode
		# If not None, `inputs` gets populated with RHS signals.
		self.inputs = inputs

	def on_Const(self, value):
		return f'{value.value}'

	def on_Signal(self, value):
		if self.inputs is not None:
			self.inputs.add(value)

		if self.mode == 'curr':
			return f'slots[{self.state.get_signal(value)}].{self.mode}'
		else:
			return f'next_{self.state.get_signal(value)}'

	def on_Operator(self, value):
		def mask(value):
			value_mask = (1 << len(value)) - 1
			return f'({value_mask:#x} & {self(value)})'

		def sign(value):
			if value.shape().signed:
				return f'sign({mask(value)}, {-1 << (len(value) - 1):#x})'
			else: # unsigned
				return mask(value)

		if len(value.operands) == 1:
			arg, = value.operands
			if value.operator == '~':
				return f'(~{mask(arg)})'
			if value.operator == '-':
				return f'(-{sign(arg)})'
			if value.operator == 'b':
				return f'bool({mask(arg)})'
			if value.operator == 'r|':
				return f'(0 != {mask(arg)})'
			if value.operator == 'r&':
				return f'({(1 << len(arg)) - 1} == {mask(arg)})'
			if value.operator == 'r^':
				# Believe it or not, this is the fastest way to compute a sideways XOR in Python.
				return f'(format({mask(arg)}, \'b\').count(\'1\') % 2)'
			if value.operator in ('u', 's'):
				# These operators don't change the bit pattern, only its interpretation.
				return self(arg)
		elif len(value.operands) == 2:
			lhs, rhs = value.operands
			if value.operator == '+':
				return f'({sign(lhs)} + {sign(rhs)})'
			if value.operator == '-':
				return f'({sign(lhs)} - {sign(rhs)})'
			if value.operator == '*':
				return f'({sign(lhs)} * {sign(rhs)})'
			if value.operator == '//':
				return f'zdiv({sign(lhs)}, {sign(rhs)})'
			if value.operator == '%':
				return f'zmod({sign(lhs)}, {sign(rhs)})'
			if value.operator == '&':
				return f'({sign(lhs)} & {sign(rhs)})'
			if value.operator == '|':
				return f'({sign(lhs)} | {sign(rhs)})'
			if value.operator == '^':
				return f'({sign(lhs)} ^ {sign(rhs)})'
			if value.operator == '<<':
				return f'({sign(lhs)} << {sign(rhs)})'
			if value.operator == '>>':
				return f'({sign(lhs)} >> {sign(rhs)})'
			if value.operator == '==':
				return f'({sign(lhs)} == {sign(rhs)})'
			if value.operator == '!=':
				return f'({sign(lhs)} != {sign(rhs)})'
			if value.operator == '<':
				return f'({sign(lhs)} < {sign(rhs)})'
			if value.operator == '<=':
				return f'({sign(lhs)} <= {sign(rhs)})'
			if value.operator == '>':
				return f'({sign(lhs)} > {sign(rhs)})'
			if value.operator == '>=':
				return f'({sign(lhs)} >= {sign(rhs)})'
		elif len(value.operands) == 3:
			if value.operator == 'm':
				sel, val1, val0 = value.operands
				return f'({sign(val1)} if {mask(sel)} else {sign(val0)})'
		raise NotImplementedError(f'Operator \'{value.operator}\' not implemented') # :nocov:

	def on_Slice(self, value):
		return f'({(1 << len(value)) - 1:#x} & ({self(value.value)} >> {value.start}))'

	def on_Part(self, value):
		offset_mask = (1 << len(value.offset)) - 1
		offset = f'({value.stride} * ({offset_mask:#x} & {self(value.offset)}))'
		return f'({(1 << value.width) - 1} & {self(value.value)} >> {offset})'

	def on_Cat(self, value):
		gen_parts = []
		offset = 0
		for part in value.parts:
			part_mask = (1 << len(part)) - 1
			gen_parts.append(f'(({part_mask:#x} & {self(part)}) << {offset})')
			offset += len(part)
		if gen_parts:
			return f'({" | ".join(gen_parts)})'
		return '0'

	def on_ArrayProxy(self, value):
		index_mask = (1 << len(value.index)) - 1
		gen_index = self.emitter.def_var('rhs_index', f'{index_mask:#x} & {self(value.index)}')
		gen_value = self.emitter.gen_var('rhs_proxy')
		if value.elems:
			if _USE_PATTERN_MATCHING:
				self.emitter.append(f'match {gen_index}:')
				with self.emitter.indent():
					for index, elem in enumerate(value.elems):
						self.emitter.append(f'case {index}:')
						with self.emitter.indent():
							self.emitter.append(f'{gen_value} = {self(elem)}')
					self.emitter.append('case _:')
					with self.emitter.indent():
						self.emitter.append(f'{gen_value} = {self(value.elems[-1])}')
			else:
				for index, elem in enumerate(value.elems):
					if index == 0:
						self.emitter.append(f'if {index} == {gen_index}:')
					else:
						self.emitter.append(f'elif {index} == {gen_index}:')
					with self.emitter.indent():
						self.emitter.append(f'{gen_value} = {self(elem)}')
				self.emitter.append('else:')
				with self.emitter.indent():
					self.emitter.append(f'{gen_value} = {self(value.elems[-1])}')

			return gen_value
		else:
			return '0'

	@classmethod
	def compile(cls, state, value, *, mode):
		emitter = _PythonEmitter()
		compiler = cls(state, emitter, mode = mode)
		emitter.append(f'result = {compiler(value)}')
		return emitter.flush()


class _LHSValueCompiler(_ValueCompiler):
	def __init__(self, state, emitter, *, rhs, outputs = None):
		super().__init__(state, emitter)
		# `rrhs` is used to translate rvalues that are syntactically a part of an lvalue, e.g.
		# the offset of a Part.
		self.rrhs = rhs
		# `lrhs` is used to translate the read part of a read-modify-write cycle during partial
		# update of an lvalue.
		self.lrhs = _RHSValueCompiler(state, emitter, mode = 'next', inputs = None)
		# If not None, `outputs` gets populated with signals on LHS.
		self.outputs = outputs

	def on_Const(self, value):
		raise TypeError # :nocov:

	def on_Signal(self, value):
		if self.outputs is not None:
			self.outputs.add(value)

		def gen(arg):
			value_mask = (1 << len(value)) - 1
			if value.shape().signed:
				value_sign = f'sign({value_mask:#x} & {arg}, {-1 << (len(value) - 1)})'
			else: # unsigned
				value_sign = f'{value_mask:#x} & {arg}'
			self.emitter.append(f'next_{self.state.get_signal(value)} = {value_sign}')
		return gen

	def on_Operator(self, value):
		if value.operator in ('u', 's'):
			return self(value.operands[0])
		raise TypeError # :nocov:

	def on_Slice(self, value):
		def gen(arg):
			width_mask = (1 << (value.stop - value.start)) - 1
			self(value.value)(
				f'({self.lrhs(value.value)} & '
				f'{~(width_mask << value.start):#x} | '
				f'(({width_mask:#x} & {arg}) << {value.start}))'
			)
		return gen

	def on_Part(self, value):
		def gen(arg):
			width_mask = (1 << value.width) - 1
			offset_mask = (1 << len(value.offset)) - 1
			offset = f'({value.stride} * ({offset_mask:#x} & {self.rrhs(value.offset)}))'
			self(value.value)(
				f'({self.lrhs(value.value)} & '
				f'~({width_mask:#x} << {offset}) | '
				f'(({width_mask:#x} & {arg}) << {offset}))'
			)
		return gen

	def on_Cat(self, value):
		def gen(arg):
			gen_arg = self.emitter.def_var('cat', arg)
			offset = 0
			for part in value.parts:
				part_mask = (1 << len(part)) - 1
				self(part)(f'({part_mask:#x} & ({gen_arg} >> {offset}))')
				offset += len(part)
		return gen

	def on_ArrayProxy(self, value):
		def gen(arg):
			index_mask = (1 << len(value.index)) - 1
			gen_index = self.emitter.def_var('index', f'{self.rrhs(value.index)} & {index_mask:#x}')
			if value.elems:
				if _USE_PATTERN_MATCHING:
					self.emitter.append(f'match {gen_index}:')
					with self.emitter.indent():
						for index, elem in enumerate(value.elems):
							self.emitter.append(f'case {index}:')
							with self.emitter.indent():
								self(elem)(arg)
						self.emitter.append('case _:')
						with self.emitter.indent():
							self(value.elems[-1])(arg)
				else:
					for index, elem in enumerate(value.elems):
						if index == 0:
							self.emitter.append(f'if {index} == {gen_index}:')
						else:
							self.emitter.append(f'elif {index} == {gen_index}:')
						with self.emitter.indent():
							self(elem)(arg)
					self.emitter.append('else:')
					with self.emitter.indent():
						self(value.elems[-1])(arg)
			else:
				self.emitter.append('pass')
		return gen


class _StatementCompiler(StatementVisitor, _Compiler):
	def __init__(self, state, emitter, *, inputs = None, outputs = None):
		super().__init__(state, emitter)
		self.rhs = _RHSValueCompiler(state, emitter, mode = 'curr', inputs = inputs)
		self.lhs = _LHSValueCompiler(state, emitter, rhs = self.rhs, outputs = outputs)

	def on_statements(self, stmts):
		for stmt in stmts:
			self(stmt)
		if not stmts:
			self.emitter.append('pass')

	def on_Assign(self, stmt):
		gen_rhs_value = self.rhs(stmt.rhs) # check for oversized value before generating mask
		gen_rhs = f'({(1 << len(stmt.rhs)) - 1:#x} & {gen_rhs_value})'
		if stmt.rhs.shape().signed:
			gen_rhs = f'sign({gen_rhs}, {-1 << (len(stmt.rhs) - 1):#x})'
		return self.lhs(stmt.lhs)(gen_rhs)

	def on_Switch(self, stmt):
		gen_test_value = self.rhs(stmt.test) # check for oversized value before generating mask
		gen_test = self.emitter.def_var('test', f'{(1 << len(stmt.test)) - 1:#x} & {gen_test_value}')
		for index, (patterns, stmts) in enumerate(stmt.cases.items()):
			gen_checks = []
			if not patterns:
				gen_checks.append('True')
			else:
				for pattern in patterns:
					if "-" in pattern:
						mask  = int(''.join('0' if b == '-' else '1' for b in pattern), 2)
						value = int(''.join('0' if b == '-' else b for b in pattern), 2)
						gen_checks.append(f'{value} == ({mask} & {gen_test})')
					else:
						value = int(pattern, 2)
						gen_checks.append(f'{value} == {gen_test}')
			if index == 0:
				self.emitter.append(f'if {" or ".join(gen_checks)}:')
			else:
				self.emitter.append(f'elif {" or ".join(gen_checks)}:')
			with self.emitter.indent():
				self(stmts)

	def on_Assert(self, stmt):
		raise NotImplementedError # :nocov:

	def on_Assume(self, stmt):
		raise NotImplementedError # :nocov:

	def on_Cover(self, stmt):
		raise NotImplementedError # :nocov:

	@classmethod
	def compile(cls, state, stmt):
		output_indexes = [state.get_signal(signal) for signal in stmt._lhs_signals()]
		emitter = _PythonEmitter()
		for signal_index in output_indexes:
			emitter.append(f'next_{signal_index} = slots[{signal_index}].next')
		compiler = cls(state, emitter)
		compiler(stmt)
		for signal_index in output_indexes:
			emitter.append(f'slots[{signal_index}].set(next_{signal_index})')
		return emitter.flush()


class _FragmentCompiler:
	def __init__(self, state):
		self.state = state

	def __call__(self, fragment):
		processes = set()

		for domain_name, domain_signals in fragment.drivers.items():
			domain_stmts = LHSGroupFilter(domain_signals)(fragment.statements)
			domain_process = PyRTLProcess(is_comb = domain_name is None)

			emitter = _PythonEmitter()
			emitter.append('def run():')
			emitter._level += 1

			if domain_name is None:
				for signal in domain_signals:
					signal_index = self.state.get_signal(signal)
					emitter.append(f'next_{signal_index} = {signal.reset}')

				inputs = SignalSet()
				_StatementCompiler(self.state, emitter, inputs = inputs)(domain_stmts)

				for input in inputs:
					self.state.add_trigger(domain_process, input)

			else:
				domain = fragment.domains[domain_name]
				clk_trigger = 1 if domain.clk_edge == 'pos' else 0
				self.state.add_trigger(domain_process, domain.clk, trigger = clk_trigger)
				if domain.rst is not None and domain.async_reset:
					rst_trigger = 1
					self.state.add_trigger(domain_process, domain.rst, trigger = rst_trigger)

				for signal in domain_signals:
					signal_index = self.state.get_signal(signal)
					emitter.append(f'next_{signal_index} = slots[{signal_index}].next')

				_StatementCompiler(self.state, emitter)(domain_stmts)

			for signal in domain_signals:
				signal_index = self.state.get_signal(signal)
				emitter.append(f'slots[{signal_index}].set(next_{signal_index})')

			# There shouldn't be any exceptions raised by the generated code, but if there are
			# (almost certainly due to a bug in the code generator), use this environment variable
			# to make backtraces useful.
			code = emitter.flush()
			if getenv('TORII_PYSIM_DUMP'):
				file = NamedTemporaryFile('w', prefix = 'torii_pysim_', delete = False)
				file.write(code)
				filename = file.name
			else:
				filename = '<string>'

			exec_locals = { 'slots': self.state.slots, **_ValueCompiler.helpers }
			exec(compile(code, filename, 'exec'), exec_locals)
			domain_process.run = exec_locals['run']

			processes.add(domain_process)

		for subfragment_index, (subfragment, subfragment_name) in enumerate(fragment.subfragments):
			if subfragment_name is None:
				subfragment_name = f'U${subfragment_index}'
			processes.update(self(subfragment))

		return processes



# SPDX-License-Identifier: BSD-2-Clause

from torii import Elaboratable, Signal, Module, Cat, Const


class UART(Elaboratable):
	'''
	Parameters
	----------
	divisor : int
		Set to ``round(clk-rate / baud-rate)``.
		E.g. ``12e6 / 115200`` = ``104``.
	'''
	def __init__(self, divisor, data_bits = 8):
		assert divisor >= 4

		self.data_bits = data_bits
		self.divisor   = divisor

		self.tx_o    = Signal()
		self.rx_i    = Signal()

		self.tx_data = Signal(data_bits)
		self.tx_rdy  = Signal()
		self.tx_ack  = Signal()

		self.rx_data = Signal(data_bits)
		self.rx_err  = Signal()
		self.rx_ovf  = Signal()
		self.rx_rdy  = Signal()
		self.rx_ack  = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()

		tx_phase = Signal(range(self.divisor))
		tx_shreg = Signal(1 + self.data_bits + 1, reset = -1)
		tx_count = Signal(range(len(tx_shreg) + 1))

		m.d.comb += self.tx_o.eq(tx_shreg[0])
		with m.If(tx_count == 0):
			m.d.comb += self.tx_ack.eq(1)
			with m.If(self.tx_rdy):
				m.d.sync += [
					tx_shreg.eq(Cat(Const(0, 1), self.tx_data, Const(1, 1))),
					tx_count.eq(len(tx_shreg)),
					tx_phase.eq(self.divisor - 1),
				]
		with m.Else():
			with m.If(tx_phase != 0):
				m.d.sync += tx_phase.eq(tx_phase - 1)
			with m.Else():
				m.d.sync += [
					tx_shreg.eq(Cat(tx_shreg[1:], Const(1, 1))),
					tx_count.eq(tx_count - 1),
					tx_phase.eq(self.divisor - 1),
				]

		rx_phase = Signal(range(self.divisor))
		rx_shreg = Signal(1 + self.data_bits + 1, reset = -1)
		rx_count = Signal(range(len(rx_shreg) + 1))

		m.d.comb += self.rx_data.eq(rx_shreg[1:-1])
		with m.If(rx_count == 0):
			m.d.comb += self.rx_err.eq(~(~rx_shreg[0] & rx_shreg[-1]))
			with m.If(~self.rx_i):
				with m.If(self.rx_ack | ~self.rx_rdy):
					m.d.sync += [
						self.rx_rdy.eq(0),
						self.rx_ovf.eq(0),
						rx_count.eq(len(rx_shreg)),
						rx_phase.eq(self.divisor // 2),
					]
				with m.Else():
					m.d.sync += self.rx_ovf.eq(1)
			with m.If(self.rx_ack):
				m.d.sync += self.rx_rdy.eq(0)
		with m.Else():
			with m.If(rx_phase != 0):
				m.d.sync += rx_phase.eq(rx_phase - 1)
			with m.Else():
				m.d.sync += [
					rx_shreg.eq(Cat(rx_shreg[1:], self.rx_i)),
					rx_count.eq(rx_count - 1),
					rx_phase.eq(self.divisor - 1),
				]
				with m.If(rx_count == 1):
					m.d.sync += self.rx_rdy.eq(1)

		return m


if __name__ == '__main__':
	uart = UART(divisor = 5)
	ports = [
		uart.tx_o, uart.rx_i,
		uart.tx_data, uart.tx_rdy, uart.tx_ack,
		uart.rx_data, uart.rx_rdy, uart.rx_err, uart.rx_ovf, uart.rx_ack
	]

	import argparse

	parser = argparse.ArgumentParser()
	p_action = parser.add_subparsers(dest = 'action')
	p_action.add_parser('simulate')
	p_action.add_parser('generate')

	args = parser.parse_args()
	if args.action == 'simulate':
		from torii.sim import Simulator, Passive

		sim = Simulator(uart)
		sim.add_clock(1e-6)

		def loopback_proc():
			yield Passive()
			while True:
				yield uart.rx_i.eq((yield uart.tx_o))
				yield
		sim.add_sync_process(loopback_proc)

		def transmit_proc():
			assert (yield uart.tx_ack)
			assert not (yield uart.rx_rdy)

			yield uart.tx_data.eq(0x5A)
			yield uart.tx_rdy.eq(1)
			yield
			yield uart.tx_rdy.eq(0)
			yield
			assert not (yield uart.tx_ack)

			for _ in range(uart.divisor * 12):
				yield

			assert (yield uart.tx_ack)
			assert (yield uart.rx_rdy)
			assert not (yield uart.rx_err)
			assert (yield uart.rx_data) == 0x5A

			yield uart.rx_ack.eq(1)
			yield
			yield uart.rx_ack.eq(0)
			yield
			assert not (yield uart.rx_rdy)

		sim.add_sync_process(transmit_proc)

		with sim.write_vcd('uart.vcd', 'uart.gtkw'):
			sim.run()

	if args.action == 'generate':
		from torii.back import verilog

		print(verilog.convert(uart, ports = ports))



# SPDX-License-Identifier: BSD-2-Clause

# If more control over clocking and resets is required, a "sync" clock domain could be created
# explicitly, which overrides the default behavior. Any other clock domains could also be
# independently created in addition to the main "sync" domain.

from torii                                     import (
	Elaboratable, Module, Signal, ClockDomain, ClockSignal
)
from torii_boards.lattice.ice40_hx1k_blink_evn import ICE40HX1KBlinkEVNPlatform

class BlinkyWithDomain(Elaboratable):
	def elaborate(self, platform) -> Module:
		clk3p3 = platform.request('clk3p3')
		led    = platform.request('led', 0)
		timer  = Signal(20)

		m = Module()
		m.domains.sync = ClockDomain()
		m.d.comb += ClockSignal().eq(clk3p3.i)
		m.d.sync += timer.eq(timer + 1)
		m.d.comb += led.o.eq(timer[-1])
		return m


if __name__ == '__main__':
	platform = ICE40HX1KBlinkEVNPlatform()
	platform.build(BlinkyWithDomain(), do_program = True)



# SPDX-License-Identifier: BSD-2-Clause

from enum           import Enum
from typing         import Iterable, Literal, Optional

from ....           import Cat, Elaboratable, Module, Record, Signal
from ....hdl.rec    import Direction
from ....util.units import log2_int
from ..memory       import MemoryMap

__all__ = (
	'Arbiter',
	'BurstTypeExt',
	'CycleType',
	'Decoder',
	'Interface',
)


class CycleType(Enum):
	''' Wishbone Registered Feedback cycle type. '''
	CLASSIC      = 0b000
	CONST_BURST  = 0b001
	INCR_BURST   = 0b010
	END_OF_BURST = 0b111


class BurstTypeExt(Enum):
	''' Wishbone Registered Feedback burst type extension. '''
	LINEAR  = 0b00
	WRAP_4  = 0b01
	WRAP_8  = 0b10
	WRAP_16 = 0b11


def _check_interface(
	addr_width: int, data_width: int, granularity: int,
	features: Iterable[Literal['rty', 'err', 'stall', 'lock', 'cti', 'bte']]
) -> None:
	if not isinstance(addr_width, int) or addr_width < 0:
		raise ValueError(f'Address width must be a non-negative integer, not {addr_width!r}')
	if data_width not in (8, 16, 32, 64):
		raise ValueError(f'Data width must be one of 8, 16, 32, 64, not {data_width!r}')
	if granularity not in (8, 16, 32, 64):
		raise ValueError(f'Granularity must be one of 8, 16, 32, 64, not {granularity!r}')
	if granularity > data_width:
		raise ValueError(f'Granularity {granularity} may not be greater than data width {data_width}')
	unknown = set(features) - { 'rty', 'err', 'stall', 'lock', 'cti', 'bte' }
	if unknown:
		raise ValueError(f'Optional signal(s) {", ".join(map(repr, unknown))} are not supported')

class Interface(Record):
	'''
	Wishbone interface.

	See the `Wishbone specification <https://opencores.org/howto/wishbone>`_ for description
	of the Wishbone signals. The ``RST_I`` and ``CLK_I`` signals are provided as a part of
	the clock domain that drives the interface.

	Note that the data width of the underlying memory map of the interface is equal to port
	granularity, not port size. If port granularity is less than port size, then the address width
	of the underlying memory map is extended to reflect that.

	Parameters
	----------
	addr_width : int
		Width of the address signal.
	data_width : int
		Width of the data signals ("port size" in Wishbone terminology).
		One of 8, 16, 32, 64.
	granularity : int
		Granularity of select signals ("port granularity" in Wishbone terminology).
		One of 8, 16, 32, 64.
	features : iter(str)
		Selects the optional signals that will be a part of this interface.
	name : str
		Name of the underlying record.

	Attributes
	----------
	The correspondence between the Torii-SoC signals and the Wishbone signals changes depending
	on whether the interface acts as an initiator or a target.

	adr : Signal(addr_width)
		Corresponds to Wishbone signal ``ADR_O`` (initiator) or ``ADR_I`` (target).
	dat_w : Signal(data_width)
		Corresponds to Wishbone signal ``DAT_O`` (initiator) or ``DAT_I`` (target).
	dat_r : Signal(data_width)
		Corresponds to Wishbone signal ``DAT_I`` (initiator) or ``DAT_O`` (target).
	sel : Signal(data_width // granularity)
		Corresponds to Wishbone signal ``SEL_O`` (initiator) or ``SEL_I`` (target).
	cyc : Signal()
		Corresponds to Wishbone signal ``CYC_O`` (initiator) or ``CYC_I`` (target).
	stb : Signal()
		Corresponds to Wishbone signal ``STB_O`` (initiator) or ``STB_I`` (target).
	we : Signal()
		Corresponds to Wishbone signal ``WE_O``  (initiator) or ``WE_I``  (target).
	ack : Signal()
		Corresponds to Wishbone signal ``ACK_I`` (initiator) or ``ACK_O`` (target).
	err : Signal()
		Optional. Corresponds to Wishbone signal ``ERR_I`` (initiator) or ``ERR_O`` (target).
	rty : Signal()
		Optional. Corresponds to Wishbone signal ``RTY_I`` (initiator) or ``RTY_O`` (target).
	stall : Signal()
		Optional. Corresponds to Wishbone signal ``STALL_I`` (initiator) or ``STALL_O`` (target).
	lock : Signal()
		Optional. Corresponds to Wishbone signal ``LOCK_O`` (initiator) or ``LOCK_I`` (target).
		torii-soc Wishbone support assumes that initiators that don't want bus arbitration to happen in
		between two transactions need to use ``lock`` feature to guarantee this. An initiator without
		the ``lock`` feature may be arbitrated in between two transactions even if ``cyc`` is kept high.
	cti : Signal()
		Optional. Corresponds to Wishbone signal ``CTI_O`` (initiator) or ``CTI_I`` (target).
	bte : Signal()
		Optional. Corresponds to Wishbone signal ``BTE_O`` (initiator) or ``BTE_I`` (target).

	'''

	def __init__(
		self, *, addr_width: int, data_width: int, granularity: int = None,
		features: Iterable[Literal['rty', 'err', 'stall', 'lock', 'cti', 'bte']] = frozenset(),
		name: Optional[str] = None
	) -> None:
		if granularity is None:
			granularity  = data_width

		_check_interface(addr_width, data_width, granularity, features)

		self.addr_width  = addr_width
		self.data_width  = data_width
		self.granularity = granularity
		self._map        = None

		features = set(features)
		layout = [
			('adr',   addr_width, Direction.FANOUT),
			('dat_w', data_width, Direction.FANOUT),
			('dat_r', data_width, Direction.FANIN),
			('sel',   data_width // granularity, Direction.FANOUT),
			('cyc',   1, Direction.FANOUT),
			('stb',   1, Direction.FANOUT),
			('we',    1, Direction.FANOUT),
			('ack',   1, Direction.FANIN),
		]
		if 'err' in features:
			layout += [('err', 1, Direction.FANIN)]
		if 'rty' in features:
			layout += [('rty', 1, Direction.FANIN)]
		if 'stall' in features:
			layout += [('stall', 1, Direction.FANIN)]
		if 'lock' in features:
			layout += [('lock',  1, Direction.FANOUT)]
		if 'cti' in features:
			layout += [('cti', CycleType,    Direction.FANOUT)]
		if 'bte' in features:
			layout += [('bte', BurstTypeExt, Direction.FANOUT)]
		super().__init__(layout, name = name, src_loc_at = 1)

	@property
	def memory_map(self) -> MemoryMap:
		if self._map is None:
			raise NotImplementedError(f'Bus interface {self!r} does not have a memory map')
		return self._map

	@memory_map.setter
	def memory_map(self, memory_map: MemoryMap) -> None:
		if not isinstance(memory_map, MemoryMap):
			raise TypeError(f'Memory map must be an instance of MemoryMap, not {memory_map!r}')
		if memory_map.data_width != self.granularity:
			raise ValueError(
				f'Memory map has data width {memory_map.data_width}, which is not the same as bus '
				f'interface granularity {self.granularity}'
			)
		granularity_bits = log2_int(self.data_width // self.granularity)
		if memory_map.addr_width != max(1, self.addr_width + granularity_bits):
			raise ValueError(
				f'Memory map has address width {memory_map.addr_width}, which is not the same as bus '
				f'interface address width {self.addr_width + granularity_bits} ({self.addr_width} '
				f'address bits + {granularity_bits} granularity bits)'
			)

		memory_map.freeze()
		self._map = memory_map


class Decoder(Elaboratable):
	'''
	Wishbone bus decoder.

	An address decoder for subordinate Wishbone buses.

	Parameters
	----------
	addr_width : int
		Address width. See :class:`Interface`.
	data_width : int
		Data width. See :class:`Interface`.
	granularity : int
		Granularity. See :class:`Interface`
	features : iter(str)
		Optional signal set. See :class:`Interface`.
	alignment : log2 of int
		Window alignment. See :class:`..memory.MemoryMap`
	name : str
		Window name. Optional.

	Attributes
	----------
	bus : :class:`Interface`
		CSR bus providing access to subordinate buses.

	'''

	def __init__(
		self, *, addr_width: int, data_width: int, granularity: Optional[int] = None,
		features: Iterable[Literal['rty', 'err', 'stall', 'lock', 'cti', 'bte']] = frozenset(),
		alignment: int = 0, name: Optional[str] = None
	) -> None:
		if granularity is None:
			granularity  = data_width

		_check_interface(addr_width, data_width, granularity, features)

		self.data_width  = data_width
		self.granularity = granularity
		self.features    = set(features)
		self.alignment   = alignment

		granularity_bits = log2_int(data_width // granularity)
		self._map        = MemoryMap(
			addr_width = max(1, addr_width + granularity_bits),
			data_width = granularity, alignment = alignment,
			name = name
		)
		self._subs       = dict()
		self._bus        = None

	@property
	def bus(self) -> Interface:
		if self._bus is None:
			self._map.freeze()
			granularity_bits = log2_int(self.data_width // self.granularity)
			self._bus = Interface(
				addr_width = self._map.addr_width - granularity_bits,
				data_width = self.data_width, granularity = self.granularity,
				features = self.features
			)
			self._bus.memory_map = self._map
		return self._bus

	def align_to(self, alignment: int) -> int:
		'''
		Align the implicit address of the next window.

		See :meth:`MemoryMap.align_to` for details.

		'''

		return self._map.align_to(alignment)

	def add(
		self, sub_bus: Interface, *, addr: Optional[int] = None, sparse: bool = False,
		extend: bool = False
	) -> tuple[int, int, int]:
		'''
		Add a window to a subordinate bus.

		The decoder can perform either sparse or dense address translation. If dense address
		translation is used (the default), the subordinate bus must have the same data width as
		the decoder; the window will be contiguous. If sparse address translation is used,
		the subordinate bus may have data width less than the data width of the decoder;
		the window may be discontiguous. In either case, the granularity of the subordinate bus
		must be equal to or less than the granularity of the decoder.

		See :meth:`MemoryMap.add_resource` for details.

		'''

		if not isinstance(sub_bus, Interface):
			raise TypeError(f'Subordinate bus must be an instance of wishbone.Interface, not {sub_bus!r}')
		if sub_bus.granularity > self.granularity:
			raise ValueError(
				f'Subordinate bus has granularity {sub_bus.granularity}, which is greater than the decoder '
				f'granularity {self.granularity}'
			)
		if not sparse:
			if sub_bus.data_width != self.data_width:
				raise ValueError(
					f'Subordinate bus has data width {sub_bus.data_width}, which is not the same as decoder data width '
					f'{self.data_width} (required for dense address translation)'
				)
		else:
			if sub_bus.granularity != sub_bus.data_width:
				raise ValueError(
					f'Subordinate bus has data width {sub_bus.data_width}, which is not the same as '
					f'subordinate bus granularity {sub_bus.granularity} (required for sparse address '
					'translation)'
				)
		for opt_output in { 'err', 'rty', 'stall' }:
			if hasattr(sub_bus, opt_output) and opt_output not in self.features:
				raise ValueError(
					f'Subordinate bus has optional output {opt_output!r}, but the decoder does not have '
					'a corresponding input'
				)

		self._subs[sub_bus.memory_map] = sub_bus
		return self._map.add_window(sub_bus.memory_map, addr = addr, sparse = sparse, extend = extend)

	def elaborate(self, platform) -> Module:
		m = Module()

		ack_fanin   = 0
		err_fanin   = 0
		rty_fanin   = 0
		stall_fanin = 0

		with m.Switch(self.bus.adr):
			for sub_map, (sub_pat, sub_ratio) in self._map.window_patterns():
				sub_bus = self._subs[sub_map]

				m.d.comb += [
					sub_bus.adr.eq(self.bus.adr << log2_int(sub_ratio)),
					sub_bus.dat_w.eq(self.bus.dat_w),
					sub_bus.sel.eq(Cat(sel.replicate(sub_ratio) for sel in self.bus.sel)),
					sub_bus.we.eq(self.bus.we),
					sub_bus.stb.eq(self.bus.stb),
				]
				if hasattr(sub_bus, 'lock'):
					m.d.comb += sub_bus.lock.eq(getattr(self.bus, 'lock', 0))
				if hasattr(sub_bus, 'cti'):
					m.d.comb += sub_bus.cti.eq(getattr(self.bus, 'cti', CycleType.CLASSIC))
				if hasattr(sub_bus, 'bte'):
					m.d.comb += sub_bus.bte.eq(getattr(self.bus, 'bte', BurstTypeExt.LINEAR))

				granularity_bits = log2_int(self.bus.data_width // self.bus.granularity)
				with m.Case(sub_pat[:-granularity_bits if granularity_bits > 0 else None]):
					m.d.comb += [
						sub_bus.cyc.eq(self.bus.cyc),
						self.bus.dat_r.eq(sub_bus.dat_r),
					]
					ack_fanin |= sub_bus.ack
					if hasattr(sub_bus, 'err'):
						err_fanin |= sub_bus.err
					if hasattr(sub_bus, 'rty'):
						rty_fanin |= sub_bus.rty
					if hasattr(sub_bus, 'stall'):
						stall_fanin |= sub_bus.stall

		m.d.comb += self.bus.ack.eq(ack_fanin)
		if hasattr(self.bus, 'err'):
			m.d.comb += self.bus.err.eq(err_fanin)
		if hasattr(self.bus, 'rty'):
			m.d.comb += self.bus.rty.eq(rty_fanin)
		if hasattr(self.bus, 'stall'):
			m.d.comb += self.bus.stall.eq(stall_fanin)

		return m


class Arbiter(Elaboratable):
	'''
	Wishbone bus arbiter.

	A round-robin arbiter for initiators accessing a shared Wishbone bus.

	Parameters
	----------
	addr_width : int
		Address width. See :class:`Interface`.
	data_width : int
		Data width. See :class:`Interface`.
	granularity : int
		Granularity. See :class:`Interface`
	features : iter(str)
		Optional signal set. See :class:`Interface`.

	Attributes
	----------
	bus : :class:`Interface`
		Shared Wishbone bus.

	'''

	def __init__(
		self, *, addr_width: int, data_width: int, granularity: Optional[int] = None,
		features: Iterable[Literal['rty', 'err', 'stall', 'lock', 'cti', 'bte']] = frozenset()
	) -> None:
		self.bus    = Interface(
			addr_width = addr_width,
			data_width = data_width,
			granularity = granularity,
			features = features
		)
		self._intrs = []

	def add(self, intr_bus: Interface) -> None:
		'''
		Add an initiator bus to the arbiter.

		The initiator bus must have the same address width and data width as the arbiter. The
		granularity of the initiator bus must be greater than or equal to the granularity of
		the arbiter.

		'''

		if not isinstance(intr_bus, Interface):
			raise TypeError(f'Initiator bus must be an instance of wishbone.Interface, not {intr_bus!r}')
		if intr_bus.addr_width != self.bus.addr_width:
			raise ValueError(
				f'Initiator bus has address width {intr_bus.addr_width}, which is not the same as arbiter '
				f'address width {self.bus.addr_width}'
			)
		if intr_bus.granularity < self.bus.granularity:
			raise ValueError(
				f'Initiator bus has granularity {intr_bus.granularity}, which is lesser than the arbiter '
				f'granularity {self.bus.granularity}'
			)
		if intr_bus.data_width != self.bus.data_width:
			raise ValueError(
				f'Initiator bus has data width {intr_bus.data_width}, which is not the same as arbiter '
				f'data width {self.bus.data_width}'
			)
		for opt_output in { 'err', 'rty' }:
			if hasattr(self.bus, opt_output) and not hasattr(intr_bus, opt_output):
				raise ValueError(
					f'Arbiter has optional output {opt_output!r}, but the initiator bus does not have '
					'a corresponding input'
				)

		self._intrs.append(intr_bus)

	def elaborate(self, platform) -> Module:
		m = Module()

		requests = Signal(len(self._intrs))
		grant    = Signal(range(len(self._intrs)))
		m.d.comb += requests.eq(Cat(intr_bus.cyc for intr_bus in self._intrs))

		bus_busy = self.bus.cyc
		if hasattr(self.bus, 'lock'):
			# If LOCK is not asserted, we also wait for STB to be de-asserted before granting bus
			# ownership to the next initiator. If we didn't, the next bus owner could receive
			# an ACK (or ERR, RTY) from the previous transaction when targeting the same
			# peripheral.
			bus_busy &= self.bus.lock | self.bus.stb

		with m.If(~bus_busy):
			with m.Switch(grant):
				for i in range(len(requests)):
					with m.Case(i):
						for pred in reversed(range(i)):
							with m.If(requests[pred]):
								m.d.sync += grant.eq(pred)
						for succ in reversed(range(i + 1, len(requests))):
							with m.If(requests[succ]):
								m.d.sync += grant.eq(succ)

		with m.Switch(grant):
			for i, intr_bus in enumerate(self._intrs):
				m.d.comb += intr_bus.dat_r.eq(self.bus.dat_r)
				if hasattr(intr_bus, 'stall'):
					intr_bus_stall = Signal(reset = 1)
					m.d.comb += intr_bus.stall.eq(intr_bus_stall)

				with m.Case(i):
					ratio = intr_bus.granularity // self.bus.granularity
					m.d.comb += [
						self.bus.adr.eq(intr_bus.adr),
						self.bus.dat_w.eq(intr_bus.dat_w),
						self.bus.sel.eq(Cat(sel.replicate(ratio) for sel in intr_bus.sel)),
						self.bus.we.eq(intr_bus.we),
						self.bus.stb.eq(intr_bus.stb),
					]
					m.d.comb += self.bus.cyc.eq(intr_bus.cyc)
					if hasattr(self.bus, 'lock'):
						m.d.comb += self.bus.lock.eq(getattr(intr_bus, 'lock', 0))
					if hasattr(self.bus, 'cti'):
						m.d.comb += self.bus.cti.eq(getattr(intr_bus, 'cti', CycleType.CLASSIC))
					if hasattr(self.bus, 'bte'):
						m.d.comb += self.bus.bte.eq(getattr(intr_bus, 'bte', BurstTypeExt.LINEAR))

					m.d.comb += intr_bus.ack.eq(self.bus.ack)
					if hasattr(intr_bus, 'err'):
						m.d.comb += intr_bus.err.eq(getattr(self.bus, 'err', 0))
					if hasattr(intr_bus, 'rty'):
						m.d.comb += intr_bus.rty.eq(getattr(self.bus, 'rty', 0))
					if hasattr(intr_bus, 'stall'):
						m.d.comb += intr_bus_stall.eq(getattr(self.bus, 'stall', ~self.bus.ack))

		return m



# SPDX-License-Identifier: BSD-2-Clause
# Reference: https://www.digilentinc.com/Pmods/Digilent-Pmod_%20Interface_Specification.pdf

from ....build import Pins, PinsN, Resource, Subsignal

__all__ = (
	'PmodDualHBridgeType6Resource',
	'PmodGPIOType1Resource',
	'PmodHBridgeType5Resource',
	'PmodSPIType2AResource',
	'PmodSPIType2Resource',
	'PmodUARTType3Resource',
	'PmodUARTType4AResource',
	'PmodUARTType4Resource',
)

def PmodGPIOType1Resource(name: str, number: int, *args, pmod: int) -> Resource:
	return Resource(
		name, number,
		Pins('1 2 3 4', dir = 'io', conn = ('pmod', pmod)),
		*args
	)


def PmodSPIType2Resource(name: str, number: int, *args, pmod: int) -> Resource:
	return Resource(
		name, number,
		Subsignal('cs',   PinsN('1', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('clk',   Pins('2', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('copi',  Pins('3', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('cipo',  Pins('4', dir = 'i', conn = ('pmod', pmod))),
		*args
	)


def PmodSPIType2AResource(name: str, number: int, *args, pmod: int) -> Resource:
	return Resource(
		name, number,
		Subsignal('cs',   PinsN('1', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('clk',   Pins('2', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('copi',  Pins('3', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('cipo',  Pins('4', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('int',   Pins('7', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('reset', Pins('8', dir = 'o', conn = ('pmod', pmod))),
		*args
	)


def PmodUARTType3Resource(name: str, number: int, *args, pmod: int) -> Resource:
	return Resource(
		name, number,
		Subsignal('cts',   Pins('1', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('rts',   Pins('2', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('rx',    Pins('3', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('tx',    Pins('4', dir = 'o', conn = ('pmod', pmod))),
		*args
	)


def PmodUARTType4Resource(name: str, number: int, *args, pmod: int) -> Resource:
	return Resource(
		name, number,
		Subsignal('cts',   Pins('1', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('tx',    Pins('2', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('rx',    Pins('3', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('rts',   Pins('4', dir = 'o', conn = ('pmod', pmod))),
		*args
	)


def PmodUARTType4AResource(name: str, number: int, *args, pmod: int) -> Resource:
	return Resource(
		name, number,
		Subsignal('cts',   Pins('1', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('tx',    Pins('2', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('rx',    Pins('3', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('rts',   Pins('4', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('int',   Pins('7', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('reset', Pins('8', dir = 'o', conn = ('pmod', pmod))),
		*args
	)


def PmodHBridgeType5Resource(name: str, number: int, *args, pmod: int) -> Resource:
	return Resource(
		name, number,
		Subsignal('dir',   Pins('1', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('en',    Pins('2', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('sa',    Pins('3', dir = 'i', conn = ('pmod', pmod))),
		Subsignal('sb',    Pins('4', dir = 'i', conn = ('pmod', pmod))),
		*args
	)


def PmodDualHBridgeType6Resource(name: str, number: int, *args, pmod: int) -> Resource:
	return Resource(
		name, number,
		Subsignal('dir',   Pins('1 3', dir = 'o', conn = ('pmod', pmod))),
		Subsignal('en',    Pins('2 4', dir = 'o', conn = ('pmod', pmod))),
		*args
	)



from torii import *


class LEDBlinker(Elaboratable):
    def elaborate(self, platform):
        m = Module()

        led = platform.request('led')

        half_freq = int(platform.default_clk_frequency // 2)
        timer = Signal(range(half_freq + 1))

        with m.If(timer == half_freq):
            m.d.sync += led.eq(~led)
            m.d.sync += timer.eq(0)
        with m.Else():
            m.d.sync += timer.eq(timer + 1)

        return m
# --- BUILD ---
from torii_boards.icestick import *


ICEStickPlatform().build(LEDBlinker(), do_program = True)



# SPDX-License-Identifier: BSD-2-Clause

from .. import Elaboratable, Module, Signal

__all__ = (
	'RoundRobin',
)


class RoundRobin(Elaboratable):
	'''
	Round-robin scheduler.

	For a given set of requests, the round-robin scheduler will
	grant one request. Once it grants a request, if any other
	requests are active, it grants the next active request with
	a greater number, restarting from zero once it reaches the
	highest one.

	Use :class:`EnableInserter` to control when the scheduler
	is updated.

	Parameters
	----------
	count : int
		Number of requests.

	Attributes
	----------
	requests : Signal(count), in
		Set of requests.
	grant : Signal(range(count)), out
		Number of the granted request. Does not change if there are no
		active requests.
	valid : Signal(), out
		Asserted if grant corresponds to an active request. De-asserted
		otherwise, i.e. if no requests are active.

	'''

	def __init__(self, *, count: int) -> None:
		if not isinstance(count, int) or count < 0:
			raise ValueError(f'Count must be a non-negative integer, not {count!r}')
		self.count    = count

		self.requests = Signal(count)
		self.grant    = Signal(range(count))
		self.valid    = Signal()

	def elaborate(self, platform) -> Module:
		m = Module()

		with m.Switch(self.grant):
			for i in range(self.count):
				with m.Case(i):
					for pred in reversed(range(i)):
						with m.If(self.requests[pred]):
							m.d.sync += self.grant.eq(pred)
					for succ in reversed(range(i + 1, self.count)):
						with m.If(self.requests[succ]):
							m.d.sync += self.grant.eq(succ)

		m.d.sync += self.valid.eq(self.requests.any())

		return m



# SPDX-License-Identifier: BSD-2-Clause

from warnings      import warn

from .lattice.ecp5 import ECP5Platform as LatticeECP5Platform

__all__ = (
	'LatticeECP5Platform',
)

warn(
	'Instead of torii.platform.vendor.lattice_ecp5, use torii.platform.vendor.lattice.ecp5',
	DeprecationWarning, stacklevel = 2
)



# SPDX-License-Identifier: BSD-2-Clause

__all__ = (
	'ps_to_sec',
	'ns_to_sec',
	'us_to_sec',
	'ms_to_sec',
	'sec_to_ps',
	'sec_to_ns',
	'sec_to_us',
	'sec_to_ms',
	'bits_for',
	'iec_size',
	'log2_int',
)

PS = 1e-12
NS = 1e-9
US = 1e-6
MS = 1e-3

def ps_to_sec(val: float) -> float:
	''' Convert the given number of picoseconds into fractional seconds '''
	return val * PS

def ns_to_sec(val: float) -> float:
	''' Convert the given number of nanoseconds into fractional seconds '''
	return val * NS

def us_to_sec(val: float) -> float:
	''' Convert the given number of microseconds into fractional seconds '''
	return val * US

def ms_to_sec(val: float) -> float:
	''' Convert the given number of milliseconds into fractional seconds '''
	return val * MS

def sec_to_ps(val: float) -> float:
	''' Convert the give number of sections into picoseconds '''
	return val / PS

def sec_to_ns(val: float) -> float:
	''' Convert the give number of sections into nanoseconds '''
	return val / NS

def sec_to_us(val: float) -> float:
	''' Convert the give number of sections into microseconds '''
	return val / US

def sec_to_ms(val: float) -> float:
	''' Convert the give number of sections into milliseconds '''
	return val / MS

def iec_size(size: int, dec: int = 2) -> str:
	''' Converts the given number of bytes to an IEC suffixed string '''
	from math import floor, log, pow

	suffixes = (
		'B'  , 'KiB', 'MiB',
		'GiB', 'TiB', 'PiB',
		'EiB', 'ZiB', 'YiB',
	)

	if size == 0:
		return '0B'

	scale = int(floor(log(size, 1024)))
	power = pow(1024, scale)
	fixed = size / power
	rem = size % power

	if rem == 0 and dec == 0:
		fixed = int(fixed)
	else:
		fixed = round(fixed, dec)

	return f'{fixed}{suffixes[scale]}'

def log2_int(n: int, need_pow2: bool = True) -> int:
	''' '''

	if n == 0:
		return 0
	r = (n - 1).bit_length()
	if need_pow2 and (1 << r) != n:
		raise ValueError(f'{n} is not a power of 2')
	return r


def bits_for(n: int, require_sign_bit: bool = False) -> int:
	''' Returns the number of bits needed to represent int ``n`` '''

	if n > 0:
		r = log2_int(n + 1, False)
	else:
		require_sign_bit = True
		r = log2_int(-n, False)
	if require_sign_bit:
		r += 1
	return r



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal
from torii.back import verilog

class Adder(Elaboratable):
	def __init__(self, width: int):
		self.a   = Signal(width)
		self.b   = Signal(width)
		self.o   = Signal(width)

	def elaborate(self, platform) -> Module:
		m = Module()
		m.d.comb += self.o.eq(self.a + self.b)
		return m


class Subtractor(Elaboratable):
	def __init__(self, width: int):
		self.a   = Signal(width)
		self.b   = Signal(width)
		self.o   = Signal(width)

	def elaborate(self, platform) -> Module:
		m = Module()
		m.d.comb += self.o.eq(self.a - self.b)
		return m


class ALU(Elaboratable):
	def __init__(self, width: int):
		self.op  = Signal()
		self.a   = Signal(width)
		self.b   = Signal(width)
		self.o   = Signal(width)

		self.add = Adder(width)
		self.sub = Subtractor(width)

	def elaborate(self, platform) -> Module:
		m = Module()
		m.submodules.add = self.add
		m.submodules.sub = self.sub
		m.d.comb += [
			self.add.a.eq(self.a),
			self.sub.a.eq(self.a),
			self.add.b.eq(self.b),
			self.sub.b.eq(self.b),
		]
		with m.If(self.op):
			m.d.comb += self.o.eq(self.sub.o)
		with m.Else():
			m.d.comb += self.o.eq(self.add.o)
		return m


if __name__ == '__main__':
	alu = ALU(width = 16)

	print(verilog.convert(alu, ports = [alu.op, alu.a, alu.b, alu.o]))
